// tuneapp.


#include "ConfigReader.h"
#include "ClassRegistry.h"

#include "xml/Parser.h"
#include "xml/Element.h"
#include "xml/Text.h"
#include "VersionDependedDocumentBuilder.h"
using namespace xml;

#include <sstream>
using namespace std;

#include <windows.h>


static int getAffectTarget(const wstring& affectString) throw(ConfigException) {
	if (affectString == L"no") {
		return Item::AFFECT_NO;
	} else if (affectString == L"system") {
		return Item::AFFECT_SYSTEM;
	} else if (affectString == L"user") {
		return Item::AFFECT_USER;
/*
	} else if (affectString == L"userPolicy") {
		return Item::AFFECT_USER_POLICY;
*/
	} else if (affectString == L"all") {
		return Item::AFFECT_ALL;
	} else {
		throw ConfigException(L"illegal <affectTarget>: " + affectString);
	}
}

static Item::PostChangedOperation getPostChangedOperation(const wstring& postChangedString) throw(ConfigException) {
	if (postChangedString == L"no") {
		return Item::PCO_NO;
	} else if (postChangedString == L"reboot") {
		return Item::PCO_REBOOT;
	} else if (postChangedString == L"logoff") {
		return Item::PCO_LOGOFF;
	} else {
		throw ConfigException(L"illegal <postChanged>: " + postChangedString);
	}
}

static inline bool isWhiteSpace(wchar_t c) {
	return (c <= 0x0020);
}


bool ConfigReader::isElementName(const xml::Node* node, const wstring& name) {
	return ((node->getType() == xml::Node::ELEMENT) && (node->getName() == name));
}


ConfigReader::ConfigReader(const wstring& baseDir)
		: baseDir(baseDir) {
}

CategoryList ConfigReader::getResults() {
	return categoryList;
}

void ConfigReader::loadDocument(const wstring& fileName, xml::Document* document) throw(IOException) {
	wstring buffer;
	utils::readFileAll(fileName, &buffer);
	wistringstream in(buffer);

	VersionDependedDocumentBuilder builder(document);
	xml::Parser(&builder).parse(in);

#ifdef _DEBUG
	document->dump();
#endif // _DEBUG
}

void ConfigReader::readConfig() throw(ConfigException) {
	set<wstring> rootFiles;
	listDir(&rootFiles, utils::concatPath(baseDir, L"root*.xml"));
	for (	set<wstring>::const_iterator iterator = rootFiles.begin();
			iterator != rootFiles.end();
			++iterator) {
		xml::Document document;
		loadDocument(utils::concatPath(baseDir, *iterator), &document);
		readConfig(&document);
	}
}

void ConfigReader::listDir(set<wstring>* files, const wstring& pattern) {
	WIN32_FIND_DATA info;
	HANDLE handle = ::FindFirstFile(pattern.c_str(), &info);
	if (handle == INVALID_HANDLE_VALUE) {
		throw IOException(L"::FindFirstFile(" + pattern + L") failed");
	}
	try {
		while (true) {
			if (wcscmp(info.cFileName, L".") && wcscmp(info.cFileName, L"..")) {
				files->insert(info.cFileName);
			}
			BOOL result = ::FindNextFile(handle, &info);
			if (!result) {
				DWORD error = ::GetLastError();
				if (error == ERROR_NO_MORE_FILES) {
					break;
				} else {
					throw IOException(L"::FindNextFile() failed");
				}
			}
		}
	} catch (...) {
		::FindClose(handle);
		throw;
	}
	::FindClose(handle);
}

void ConfigReader::readConfig(xml::Document* document) throw(ConfigException) {
	for (	xml::Node::const_iterator children = document->begin();
			children != document->end();
			++children) {
		xml::Node* child = *children;
		if (isElementName(child, L"groups")) {
			readCategories(child);
			return;
		}
	}
	throw ConfigException(L"no groups element");
}

void ConfigReader::readCategories(xml::Node* root) throw(ConfigException) {
	for (	xml::Node::const_iterator children = root->begin();
			children != root->end();
			++children) {
		xml::Node* child = *children;
		if (isElementName(child, L"group")) {
			readCategory(child);
		}
	}
}

Category* ConfigReader::readCategory(xml::Node* root) throw(ConfigException) {
	const wstring& src = root->getAttribute(L"src");
	if (src != NULL_STRING) {
		Category* category = includeCategory(src);
		if (root->hasAttribute(L"expand")) {
			category->setExpand(parseBool(root->getAttribute(L"expand")));
		}
		for (	xml::Node::const_iterator subGroups = root->begin();
				subGroups != root->end();
				++subGroups) {
			xml::Node* subGroup = *subGroups;
			if (isElementName(subGroup, L"group")) {
				Category* subCategory = readCategory(subGroup);
				if (subCategory->getParent() == NULL) {
					subCategory->setParent(category);
					category->addSubCategory(subCategory);
				}
			}
		}
		return category;
	} else {
		return readCategoryImpl(root);
	}
}

Category* ConfigReader::includeCategory(const wstring& src) throw(ConfigException) {
	xml::Document document;
	loadDocument(utils::concatPath(baseDir, src), &document);
	return readCategoryConfig(&document);
}

Category* ConfigReader::readCategoryConfig(xml::Document* document) throw(ConfigException) {
	for (	xml::Node::const_iterator children = document->begin();
			children != document->end();
			++children) {
		xml::Node* child = *children;
		if (isElementName(child, L"group")) {
			return readCategoryImpl(child);
		}
	}
	throw ConfigException(L"no group element");
}

Category* ConfigReader::getCategory(const wstring& id, const wstring& className) {
	Category* category;
	for (	CategoryList::iterator iterator = categoryList.begin();
			iterator != categoryList.end();
			++iterator) {
		category = *iterator;
		if (category->getID() == id) {
			if (className != NULL_STRING) {
				throw ConfigException(L"duplicate id with class: " + id);
			}
			return category;
		}
	}
	if (className == NULL_STRING) {
		category = new Category();
	} else {
		try {
			category = categoryClassRegistry.create(className);
		} catch (IllegalArgumentException) {
			throw ConfigException(className);
		}
	}
	category->setID(id);
	categoryList.push_back(category);
	return category;
}

Category* ConfigReader::readCategoryImpl(xml::Node* root) throw(ConfigException) {
	Category* category;
	const wstring& id = root->getAttribute(L"id");
	if ((id == NULL_STRING) || id.empty()) {
		throw ConfigException(L"no id specified on category defenition (group element)");
	}
	const wstring& className = root->getAttribute(L"class");
	category = getCategory(id, className);
	{
		for (	xml::Node::const_iterator children = root->begin();
				children != root->end();
				++children) {
			xml::Node* child = *children;
			if (isElementName(child, L"params")) {
				category->setParams(child);
			}
		}
	}
	{
		for (	xml::Node::const_iterator children = root->begin();
				children != root->end();
				++children) {
			xml::Node* child = *children;
			if (child->getType() == xml::Node::ELEMENT) {
				const wstring& childName = child->getName();
				if (childName == L"name") {
					category->setName(child->getText());
				} else if (childName == L"help") {
					category->setHelpText(getHelpText(child));
				} else if (childName == L"group") {
					ItemGroup* group = readItemGroup(child);
					group->setParent(category);
					category->addGroup(group);
				} else if (childName == L"item") {
					ItemGroup* group = readSigleItemGroup(child);
					group->setParent(category);
					category->addGroup(group);
				}
			}
		}
	}
	return category;
}

ItemGroup* ConfigReader::readItemGroup(xml::Node* root) throw(ConfigException) {
	ItemGroup* group = new ItemGroup();
	const wstring& id = root->getAttribute(L"id");
	if ((id == NULL_STRING) || id.empty()) {
		throw ConfigException(L"no id specified in group element");
	}
	group->setID(id);
	for (	xml::Node::const_iterator children = root->begin();
			children != root->end();
			++children) {
		xml::Node* child = *children;
		if (child->getType() == xml::Node::ELEMENT) {
			const wstring& childName = child->getName();
			if (childName == L"name") {
				group->setName(child->getText());
			} else if (childName == L"help") {
				group->setHelpText(getHelpText(child));
			} else if (childName == L"shortHelp") {
				group->setShortHelpText(child->getText());
			} else if (childName == L"item") {
				Item* item = readItem(child);
				item->setParent(group);
				group->addItem(item);
			}
		}
	}
	return group;
}

ItemGroup* ConfigReader::readSigleItemGroup(xml::Node* root) throw(ConfigException) {
	ItemGroup* group = new ItemGroup();
	Item* item = readItem(root);
	item->setParent(group);
	group->setID(item->getID());
	group->setName(item->getName());
	group->setHelpText(item->getHelpText());
	group->setShortHelpText(item->getShortHelpText());
	group->addItem(item);
	return group;
}

Item* ConfigReader::readItem(xml::Node* root) throw(ConfigException) {
	const wstring& className = root->getAttribute(L"class");
	if ((className == NULL_STRING) || className.empty()) {
		throw ConfigException(L"no class attribute specified in item element");
	}
	Item* item;
	try {
		item = itemClassRegistry.create(className);
	} catch (IllegalArgumentException) {
		throw ConfigException(className);
	}
	const wstring& id = root->getAttribute(L"id");
	if ((id == NULL_STRING) || id.empty()){
		throw ConfigException(L"no id specified in item element");
	}
	item->setID(id);
	for (	xml::Node::const_iterator children = root->begin();
			children != root->end();
			++children) {
		xml::Node* child = *children;
		if (child->getType() == xml::Node::ELEMENT) {
			const wstring& childName = child->getName();
			if (childName == L"name") {
				item->setName(child->getText());
			} else if (childName == L"params") {
				item->setParams(child);
			} else if (childName == L"affectTarget") {
				item->setAffectTarget(getAffectTarget(child->getText()));
			} else if (childName == L"postChanged") {
				item->setPostChangedOperation(getPostChangedOperation(child->getText()));
			} else if (childName == L"help") {
				item->setHelpText(getHelpText(child));
			}
			// short help ...?
		}
	}
	return item;
}


// help


static wstring eraceCountinuouslySpace(const wstring& source) {
	wstring buffer;
	buffer.reserve(source.length());
	bool postSpace = false;
	for (int i = 0; i < source.length(); i++) {
		wchar_t c = source.at(i);
		if (isWhiteSpace(c)) {
			if (!postSpace) {
				buffer += L' ';
				postSpace = true;
			}
		} else {
			buffer += c;
			postSpace = false;
		}
	}
	return buffer;
}

static wstring replaceBrToLn(const wstring& source) {
	wstring buffer;
	buffer.reserve(source.length());
	wstring::size_type start = 0;
	wstring::size_type br;
	while ((br = source.find(L"<br>", start)) != wstring::npos) {
		buffer += trim(source.substr(start, br - start));
		buffer += L"\r\n";
		start = br + 4;
	}
	buffer += trim(source.substr(start));
	return buffer;
}

wstring ConfigReader::getHelpText(xml::Node* root) {
	if ((root->getChildCount() == 1) && ((*root->begin())->getType() == Node::Type::TEXT)) {
		// traditional (CDARA)
		wstring source(eraceCountinuouslySpace(root->getText()));
		return replaceBrToLn(source);
	} else {
		wstring result;
		getHelpText(&result, root);
		return result;
	}
}

void ConfigReader::getHelpText(wstring* result, xml::Node* root) {
	for (	xml::Node::const_iterator children = root->begin();
			children != root->end();
			++children) {
		xml::Node* child = *children;
		switch (child->getType()) {
		case Node::Type::TEXT:
			*result += eraceCountinuouslySpace(child->getText());
			break;
		case Node::Type::ELEMENT:
			if (child->getName() == L"br") {
				*result += L'\r';
			} else {
				getHelpText(result, child);
			}
			break;
		default:
			getHelpText(result, child);
			break;
		}
	}
}
