// tuneapp


#include "TuneApp.h"
#include "ClassRegistry.h"
#include "ConfigReader.h"

#include "utils.h"

#include "awt/Button.h"
#include "awt/Panel.h"
#include "awt/BorderedPanel.h"
#include "awt/CardLayout.h"
#include "awt/FileDialog.h"
#include "awt/Toolkit.h"
#include "awt/FitLayout.h"
#include "ui/UsefulTree.h"

#include <windows.h>

#include <assert.h>

#include "resource.h"


const wstring TuneApp::APP_NAME(L"tuneapp");
const wstring TuneApp::VERSION(L"1.3.2");
const wstring TuneApp::VERSION_SUFIX(L"production release");

void createArgs(vector<wstring>* args);


/**
 *	Application entry point.
 */
int WINAPI WinMain(
		HINSTANCE hInstance, HINSTANCE hPrevInstance,
		LPSTR lpCmdLine, int nCmdShow) {
	if (::CoInitialize(NULL) != S_OK) {
		::MessageBox(NULL,
				L"CoInitialize() failed",
				L"tuneapp",
				MB_OK | MB_ICONERROR | MB_TASKMODAL);
		return 1;
	}
	awt::Toolkit::init(hInstance, hPrevInstance, lpCmdLine, nCmdShow);
	utils::init();
	registerClasses();

	TuneApp* application;
	try {
		vector<wstring> args;
		createArgs(&args);
		application = new TuneApp(args);
	} catch (Exception& e) {
		::MessageBox(NULL,
				e.getMessage().c_str(),
				L"tuneapp",
				MB_OK | MB_ICONERROR | MB_TASKMODAL);
		return 1;
	} catch (exception& e) {
		::MessageBoxA(NULL,
				e.what(),
				"tuneapp",
				MB_OK | MB_ICONERROR | MB_TASKMODAL);
		return 1;
/*
	} catch (...) {
		::MessageBoxA(NULL,
				"unknown initialozation error",
				"tuneapp",
				MB_OK | MB_ICONERROR | MB_TASKMODAL);
		return 1;
*/
	}
	// application->setVisible(nCmdShow);
	application->setVisible(true);
#ifdef _DEBUG
	application->debug();
#endif

	int result = application->getToolkit()->messageLoop();

	delete application;

	CoUninitialize();

	return result;
}

void createArgs(vector<wstring>* args) {
	int argc;
	wchar_t** argv = CommandLineToArgvW(GetCommandLine(), &argc);
	try {
		args->resize(argc - 1);
		for (int i = 1; i < argc; i++) {
			args->at(i - 1) = argv[i];
		}
	} catch (...) {
		::GlobalFree(argv);
		throw;
	}
	::GlobalFree(argv);
}

/**
 *	Application
 */
TuneApp::TuneApp(const vector<wstring>& args)
		: itemsPath(L"items"), affect(Item::AFFECT_ALL) {
	parseArgs(args);
	readConfig();
	createComponents();
	showCategory(categories.at(0));
}

void TuneApp::parseArgs(const vector<wstring>& args) throw(ConfigException) {
	wstring key = NULL_STRING;
	for (	vector<wstring>::const_iterator iterator = args.begin();
			iterator != args.end();
			++iterator) {
		const wstring& arg = *iterator;
		if (arg.at(0) == '/') {
			if (key == NULL_STRING) {
				key = arg;
			} else {
				setArgument(key, NULL_STRING);
				key = arg;
			}
		} else {
			if (key == NULL_STRING) {
				throw ConfigException(L"invalid argument: " + arg);
			} else {
				setArgument(key, arg);
				key = NULL_STRING;
			}
		}
	}
	if (key != NULL_STRING) {
		setArgument(key, NULL_STRING);
	}
}

void TuneApp::setArgument(const wstring& key, const wstring& value) throw(ConfigException) {
	if (key == L"/items") {
		if (value == NULL_STRING) {
			throw ConfigException(L"/items needs to specify the directory");
		}
		itemsPath = value;
	} else {
		throw ConfigException(L"invalid argument: " + key + L" " + value);
	}
}

/**
 *	Read configuration files.
 *	from item/ directory or specified one via command line.
 */
void TuneApp::readConfig() {
	ConfigReader configReader(itemsPath);
	configReader.readConfig();
	categories = configReader.getResults();
}


// UI.


/**
 *	Create UI components.
 */
void TuneApp::createComponents() {
	setTitle(getToolkit()->getStringResource(IDS_TITLE));

	categoryTree = createCategoryTree();
	affectChoice = createAffectSelector();
	customizerContainer = createCustomizerContainer();
	helpTextArea = createHelpTextArea();
	awt::Container* buttonPane = createButtonPane();
	statusBar = createStatusBar();

	awt::Container* categorySelectorPane = new awt::Panel();
	categorySelectorPane->setLayout(new awt::BorderLayout());
	categorySelectorPane->add(createCategoriesTitle(), awt::BorderLayout::NORTH);
	categorySelectorPane->add(categoryTree, awt::BorderLayout::CENTER);
	categorySelectorPane->add(affectChoice, awt::BorderLayout::SOUTH);

	awt::Container* mainPane = new awt::Panel();
	mainPane->setLayout(rootLayout = new awt::BorderLayout());
	mainPane->add(categorySelectorPane, awt::BorderLayout::WEST);
	mainPane->add(customizerContainer, awt::BorderLayout::CENTER);

	awt::Container* helpSpliter = new awt::Panel();
	helpSpliter->setLayout(helpLayout = new awt::BorderLayout());
	helpSpliter->add(mainPane, awt::BorderLayout::NORTH);
	helpSpliter->add(helpTextArea, awt::BorderLayout::CENTER);
	helpSpliter->add(buttonPane, awt::BorderLayout::SOUTH);

	setLayout(statusLayout = new awt::BorderLayout());
	add(helpSpliter, awt::BorderLayout::CENTER);
	add(statusBar, awt::BorderLayout::SOUTH);

	addWindowListener(this);
	getToolkit()->addKeyListener(this);

	// setResizable(true);
	pack();
	setMinimumSize(getSize());

	loadIcon();
}

awt::Choice* TuneApp::createAffectSelector() {
	awt::Choice* choice = new awt::Choice();
	choice->add(getToolkit()->getStringResource(IDS_AFFECT_ALL));
	choice->add(getToolkit()->getStringResource(IDS_AFFECT_SYSTEM));
	choice->add(getToolkit()->getStringResource(IDS_AFFECT_USER));
	choice->select(0);
	choice->addItemListener(new awt::item_mem_fun<TuneApp>(this, affectSelectorChanged));
	return choice;
}

awt::Component* TuneApp::createCategoriesTitle() {
	awt::Label* label = new awt::Label(getToolkit()->getStringResource(IDS_CATEGORY));
	awt::Container* container = new awt::Panel();
	container->setLayout(awt::FitLayout::sharedInstance);
	int hgap = ::GetSystemMetrics(SM_CXEDGE) + ::GetSystemMetrics(SM_CXBORDER);
	int vgap = ::GetSystemMetrics(SM_CYEDGE) + ::GetSystemMetrics(SM_CYBORDER);
	container->setInsets(awt::Insets(vgap, 0, vgap, 0));
	container->add(label);
	return container;
}

awt::Tree* TuneApp::createCategoryTree() {
	awt::Tree* tree = new UsefulTree();
	bool showRootHandle = false;
	for (	CategoryList::const_iterator iterator = categories.begin();
			iterator != categories.end();
			++iterator) {
		if (!(*iterator)->getSubCategoryList().empty()) {
			showRootHandle = true;
			break;
		}
	}
	tree->setShowsRootHandles(showRootHandle);
	updateCategoryTree(tree);
	tree->addItemListener(new awt::item_mem_fun<TuneApp>(this, categoryTreeChanged));
	return tree;
}

awt::Container* TuneApp::createCustomizerContainer() {
	awt::Container* container = new awt::Panel();
	container->setLayout(new awt::CardLayout());
	groupsLayout = new awt::GridFlowLayout(0, 1);
	groupsLayout->setValign(awt::GridFlowLayout::TOP);
	addCategoriesView(categories, container);
	return container;
}

void TuneApp::addCategoriesView(const CategoryList& categoryList, awt::Container* container) {
	for (	CategoryList::const_iterator iterator = categoryList.begin();
			iterator != categoryList.end();
			++iterator) {
		Category* category = *iterator;
		awt::Container* categoryContainer;
		if (ItemGroupSelector* selector = dynamic_cast<ItemGroupSelector*>(category)) {
			categoryContainer = createGroupSelectorContainer(selector);
		} else {
			categoryContainer = createCategoryContainer(category);
		}
		container->add(categoryContainer, awt::CardLayout::Constraints(category->getID()));
		viewMap[category] = categoryContainer;
		addCategoriesView(category->getSubCategoryList(), container);
	}
}

awt::Container* TuneApp::createGroupSelectorContainer(ItemGroupSelector* selector) {
	vector<wstring> selections;

	awt::Container* container = new awt::Panel();
	awt::Choice* choice = new awt::Choice();
	awt::Container* choiceContainer = new awt::Panel();
	choiceContainer->setLayout(utils::leftFlowLayout);
	choiceContainer->add(new awt::Label(selector->getLabel()));
	choiceContainer->add(choice);
	(new HelpKicker(this, selector))->installTo(choiceContainer);
	awt::Container* selectionContainer = new awt::Panel();
	selectionContainer->setLayout(new awt::CardLayout());
	container->setLayout(groupsLayout);
	container->add(createCaption(selector->getName()));
	container->add(choiceContainer);
	container->add(selectionContainer);

	for (	vector<wstring>::const_iterator keys = selector->keys().begin();
			keys != selector->keys().end();
			++keys) {
		const wstring& key = *keys;
		selections.push_back(key);
		const wstring& label = selector->getLabel(key);
		choice->add(label);
		awt::Container* groupsContainer = new awt::Panel();
		groupsContainer->setLayout(groupsLayout);
		const ItemGroupList& groupList = selector->getItemGroupList(key);
		for (	ItemGroupList::const_iterator groups = groupList.begin();
				groups != groupList.end();
				++groups) {
			ItemGroup* group = *groups;
			awt::Component* view = createGroupView(group);
			(new HelpKicker(this, group))->installTo(view);
			groupsContainer->add(view);
		}
		selectionContainer->add(groupsContainer, awt::CardLayout::Constraints(key));
	}

	choice->select(0);
	choice->addItemListener(new ItemGroupSelectorListener(
			this,
			selections, choice, selectionContainer) );

	return container;
}

awt::Container* TuneApp::createCategoryContainer(Category* category) {
	awt::Container* container = new awt::Panel();
	container->setLayout(groupsLayout);
	container->add(createCaption(category->getName()));
	const ItemGroupList& groupList = category->getItemGroupList();
	for (	ItemGroupList::const_iterator iterator = groupList.begin();
			iterator != groupList.end();
			++iterator) {
		ItemGroup* group = *iterator;
		awt::Component* view = createGroupView(group);
		(new HelpKicker(this, group))->installTo(view);
		container->add(view);
	}
	return container;
}

awt::Component* TuneApp::createCaption(const wstring& text) {
	awt::Label* label = new awt::Label(text);
	label->setForeground(::GetSysColor(COLOR_CAPTIONTEXT));
	label->setBackground(::GetSysColor(COLOR_ACTIVECAPTION));
	awt::Container* container = new awt::Panel();
	container->setLayout(awt::FitLayout::sharedInstance);
	int hgap = ::GetSystemMetrics(SM_CXEDGE) + ::GetSystemMetrics(SM_CXBORDER);
	int vgap = ::GetSystemMetrics(SM_CYEDGE) + ::GetSystemMetrics(SM_CYBORDER);
	container->setInsets(awt::Insets(vgap, hgap, vgap, hgap));
	container->setBackground(::GetSysColor(COLOR_ACTIVECAPTION));
	container->add(label);
	return container;
}

awt::Component* TuneApp::createGroupView(ItemGroup* group) {
	awt::BorderedPanel* container = new awt::BorderedPanel(group->getName());
	viewMap[group] = container;
	if (group->countItems() == 1) {
		container->setLayout(utils::rowsGridLayout);
		for (	ItemGroup::const_iterator iterator = group->begin();
				iterator != group->end();
				++iterator) {
			Item* item = *iterator;
			awt::Component* view = item->getViewComponent();
			viewMap[item] = view;
			container->add(view);
		}
	} else {
		container->setLayout(utils::c2GridFlowLayout);
		for (	ItemGroup::const_iterator iterator = group->begin();
				iterator != group->end();
				++iterator) {
			Item* item = *iterator;
			awt::Component* view = item->getViewComponent();
			viewMap[item] = view;
			container->add(new awt::Label(item->getName()));
			container->add(view);
		}
	}
	return container;
}

awt::TextArea* TuneApp::createHelpTextArea() {
	awt::TextArea* helpTextArea = new awt::TextArea(L"", 8, 100);
	helpTextArea->setEditable(false);
	helpTextArea->setLineSpace(2);
	return helpTextArea;
}

awt::Container* TuneApp::createButtonPane() {
	awt::Button* okButton = new awt::Button(getToolkit()->getStringResource(IDS_OK));
	awt::Button* cancelButton = new awt::Button(getToolkit()->getStringResource(IDS_CANCEL));
	awt::Button* saveButton = new awt::Button(getToolkit()->getStringResource(IDS_SAVE));

	awt::Container* c0 = new awt::Panel();
	c0->setLayout(buttonsLayout1 = new awt::GridLayout(1, 0));
	c0->add(okButton);
	c0->add(cancelButton);

	awt::Container* container = new awt::Panel();
	container->setLayout(buttonsLayout0 = new awt::FlowLayout(awt::FlowLayout::CENTER));
	container->add(c0);
	container->add(saveButton);

	okButton->addActionListener(new awt::action_mem_fun<TuneApp>(this, ok));
	cancelButton->addActionListener(new awt::action_mem_fun<TuneApp>(this, cancel));
	saveButton->addActionListener(new awt::action_mem_fun<TuneApp>(this, save));

	return container;
}

awt::Label* TuneApp::createStatusBar() {
	return new awt::Label(APP_NAME + L"/" + VERSION);
}

void TuneApp::addNotify() {
	Frame::addNotify();

	TEXTMETRIC metrics;
	getTextMetrics(NULL_STRING, NULL, &metrics);

	rootLayout->setHgap(metrics.tmAveCharWidth);
	groupsLayout->setVgap(metrics.tmHeight / 2);
	helpLayout->setVgap(metrics.tmHeight / 2);
	buttonsLayout0->setHgap(metrics.tmAveCharWidth * 2);
	buttonsLayout1->setHgap(metrics.tmAveCharWidth);
	// statusLayout->setVgap(metrics.tmHeight / 2);

	utils::c2GridFlowLayout->setHgap(metrics.tmAveCharWidth);
	utils::c2GridFlowLayout->setVgap(metrics.tmHeight / 2);
	utils::leftFlowLayout->setHgap(metrics.tmAveCharWidth);
}

/**
 *	load window icon from "tuneapp.ico"
 */
void TuneApp::loadIcon() {
	static const wchar_t* fileName = L"tuneapp.ico";
	HICON smallIcon = static_cast<HICON>(::LoadImage(
			NULL, fileName, IMAGE_ICON,
			::GetSystemMetrics(SM_CXSMICON), ::GetSystemMetrics(SM_CYSMICON),
			LR_LOADFROMFILE) );
	HICON largeIcon = static_cast<HICON>(::LoadImage(
			NULL, fileName, IMAGE_ICON,
			::GetSystemMetrics(SM_CXICON), ::GetSystemMetrics(SM_CYICON),
			LR_LOADFROMFILE) );
	if ((smallIcon == NULL) || (largeIcon == NULL)) {
		return;
	}
	setIcon(smallIcon, largeIcon);
}


// UI listeners implementation.


/**
 *	Category tree listener.
 */
void TuneApp::categoryTreeChanged() {
	if (eventLocked()) return;
	awt::Tree::Node* node = categoryTree->getSelectedNode();
	if (node != NULL) {
		Category* category = static_cast<Category*>(node->getValue());
		assert (category != NULL);
		showCategory(category);
	}
}

/**
 *	Affect selector listener.
 */
void TuneApp::affectSelectorChanged() {
	static int affectList[] = {
		Item::AFFECT_ALL,
		Item::AFFECT_SYSTEM,
		Item::AFFECT_USER
	};
	int affect = affectList[affectChoice->getSelectedIndex()];
	setAffect(affect);
}

/**
 *	ItemGroupSelector listener.
 */
TuneApp::ItemGroupSelectorListener::ItemGroupSelectorListener(
		TuneApp* tuneapp,
		const vector<wstring>& keys,
		awt::Choice* choice, awt::Container* selectionContainer)
		: tuneapp(tuneapp), keys(keys), choice(choice), selectionContainer(selectionContainer) {
}

void TuneApp::ItemGroupSelectorListener::itemStateChanged(awt::ItemEvent& event) {
	const wstring& key = keys.at(choice->getSelectedIndex());
	awt::CardLayout* layout = dynamic_cast<awt::CardLayout*>(selectionContainer->getLayout());
	assert (layout != NULL);
	layout->show(selectionContainer, key);
}

/**
 *	Window listener.
 */
void TuneApp::windowClosing(awt::WindowEvent& event) {
	cancel();
}

/**
 *	Key listener.
 */
void TuneApp::keyPressed(awt::KeyEvent& event) {
	if (event.getModifiers() == awt::KeyEvent::CTRL_MASK) {
		switch (event.getKeyCode()) {
		case 'S':
			save();
			event.consume();
			break;
		case 'M':
			playMp3();
			event.consume();
			break;
		}
	} else {
		switch (event.getKeyCode()) {
		case VK_PRIOR: // VK_PAGE_UP
			showCategoryInc(-1);
			event.consume();
			break;
		case VK_NEXT: // VK_PAGE_DOWN
			showCategoryInc(+1);
			event.consume();
			break;
		case VK_F1:
			about();
			event.consume();
			break;
		case VK_RETURN: // VK_ENTER
			event.consume();
			ok();
			break;
		case VK_ESCAPE:
			cancel();
			event.consume();
			break;
		}
	}
}

void TuneApp::keyReleased(awt::KeyEvent& event) {}

void TuneApp::keyTyped(awt::KeyEvent& event) {}


// HelpKicker implementation.


TuneApp::HelpKicker::HelpKicker(TuneApp* tuneapp, Customizer* customizer)
		: tuneapp(tuneapp), customizer(customizer),
		entered(false), popupScheduled(false) {
}

UINT TuneApp::HelpKicker::getMouseHoverTime() {
	// return ::GetDoubleClickTime();
	return 800;
}

void TuneApp::HelpKicker::installDispatcherTo(awt::Component* component) {
	component->addMouseListener(this);
	component->addFocusListener(this);
}

void TuneApp::HelpKicker::uninstallDispatcherFrom(awt::Component* component) {
	component->removeMouseListener(this);
	component->removeFocusListener(this);
}

void TuneApp::HelpKicker::focusGained(awt::FocusEvent& event) {
	tuneapp->showHelp(customizer);
}

void TuneApp::HelpKicker::mousePressed(awt::MouseEvent& event) {
	tuneapp->showHelp(customizer);
}

void TuneApp::HelpKicker::mouseEntered(awt::MouseEvent& event) {
	entered = true;
	if (!popupScheduled) {
		popupScheduled = true;
		tuneapp->setTimer(getMouseHoverTime(), this);
	}
}

void TuneApp::HelpKicker::mouseExited(awt::MouseEvent& event) {
	entered = false;
}

void TuneApp::HelpKicker::actionPerformed(awt::ActionEvent& event) {
	if (entered) {
		tuneapp->showHelp(customizer);
	}
	popupScheduled = false;
}


// Event lock operations.


bool TuneApp::eventLocked() {
	return (eventLockCount > 0);
}

void TuneApp::lockEvent() {
	eventLockCount++;
}

void TuneApp::unlockEvent() {
	eventLockCount--;
}


// Change affect target.


static bool isEnabledAffect(int affect, int mask) {
	return (affect & mask) != 0;
}

void TuneApp::setAffect(int affect) {
	if (this->affect == affect) return;
	this->affect = affect;
	awt::Tree::Node* oldSelected = categoryTree->getSelectedNode();
	Category* oldCategory;
	if (oldSelected == NULL) {
		oldCategory = NULL;
	} else {
		oldCategory = static_cast<Category*>(oldSelected->getValue());
	}
	updateViewAffection(categories);
	updateCategoryTree();
	if (oldCategory != NULL) {
		showCategory(oldCategory);
	}
}

void TuneApp::updateViewAffection(const CategoryList& categoryList) {
	bool enabled;
	awt::Component* view;
	for (	CategoryList::const_iterator categories = categoryList.begin();
			categories != categoryList.end();
			++categories) {
		Category* category = *categories;
		const ItemGroupList& groupList = category->getItemGroupList();
		for (	ItemGroupList::const_iterator groups = groupList.begin();
				groups != groupList.end();
				++groups) {
			ItemGroup* group = *groups;
			enabled = isEnabledAffect(group->getAffectTarget(), affect);
			view = utils::find_ptr<Customizer*, awt::Component*>(viewMap, group);
			assert (view != NULL);
			view->setVisible(enabled);
			for (	ItemGroup::const_iterator items = group->begin();
					items != group->end();
					++items) {
				Item* item = *items;
				enabled = isEnabledAffect(item->getAffectTarget(), affect);
				view = utils::find_ptr<Customizer*, awt::Component*>(viewMap, item);
				assert (view != NULL);
				view->setVisible(enabled);
			}
		}
		updateViewAffection(category->getSubCategoryList());
	}
}

void TuneApp::updateCategoryTree() {
	updateCategoryTree(categoryTree);
}

void TuneApp::updateCategoryTree(awt::Tree* tree) {
	tree->removeAll();
	updateCategoryTreeAffection(tree, tree->getRootNode(), categories);
}

void TuneApp::updateCategoryTreeAffection(awt::Tree* tree, awt::Tree::Node* parent, const CategoryList& categoryList) {
	for (	CategoryList::const_iterator categories = categoryList.begin();
			categories != categoryList.end();
			++categories) {
		Category* category = *categories;
		if (isEnabledAffect(category->getAffectTarget(), affect)) {
			awt::Tree::Node* node = tree->addNode(parent, category->getName(), category);
			tree->setExpanded(node, category->isExpand());
			updateCategoryTreeAffection(tree, node, category->getSubCategoryList());
		}
	}
}


// Category operation.


void TuneApp::showCategory(Category* category) throw(IllegalArgumentException) {
	awt::Tree::Node* node = getTreeNodeFor(category);
	if (node == NULL) {
		awt::Tree::Node* rootNode = categoryTree->getRootNode();
		if (rootNode->getChildCount() == 0) return;
		node = rootNode->getChild(0);
		category = static_cast<Category*>(node->getValue());
	}
	lockEvent();
	try {
		categoryTree->select(node);
		awt::CardLayout* layout = dynamic_cast<awt::CardLayout*>(
				customizerContainer->getLayout());
		assert (layout != NULL);
		layout->show(customizerContainer, category->getID());
	} catch (...) {
		unlockEvent();
		throw;
	}
	unlockEvent();
	showHelp(category);
}

void TuneApp::showCategoryInc(int step) {
	awt::Tree::Node* node = categoryTree->getSelectedNode();
	if (node == NULL) return;
	if (step == 1) {
		node = categoryTree->getNextNode(node);
	} else if (step == -1) {
		node = categoryTree->getPreviousNode(node);
	} else {
		assert (false);
	}
	if (node == NULL) return;
	Category* category = static_cast<Category*>(node->getValue());
	assert (category != NULL);
	showCategory(category);
	awt::Component* view = utils::find_ptr<Customizer*, awt::Component*>(viewMap, category);
	assert (view != NULL);
	view->requestFocus();
}

awt::Tree::Node* TuneApp::getTreeNodeFor(Category* category) throw(IllegalArgumentException) {
	CategoryList reversePath;
	for (Category* c = category; c != NULL; c = c->getParent()) {
		reversePath.push_back(c);
	}
	awt::Tree::Node* node = categoryTree->getRootNode();
	for (	CategoryList::reverse_iterator iterator = reversePath.rbegin();
			iterator != reversePath.rend();
			++iterator) {
		node = categoryTree->findValue(node, *iterator);
		if (node == NULL) return NULL;
	}
	return node;
}


// Help.


void TuneApp::showHelp(const Customizer* target) {
	setHelpText(target->getHelpText());
	// setStatus(target->getShortHelpText());
}

void TuneApp::setHelpText(const wstring& helpText) {
	if (helpTextArea->getText() != helpText) {
		helpTextArea->setCaretPosition(0);
		helpTextArea->setText(helpText);
	}
}

void TuneApp::setStatus(const wstring& status) {
	if (status.empty() || (status == NULL_STRING)) {
		statusBar->setText(APP_NAME + L"/" + VERSION);
	} else {
		statusBar->setText(status);
	}
}


// Button actions.


void TuneApp::ok() {
	apply();
	dispose();
	::PostQuitMessage(0);
}

void TuneApp::apply() {
	Applyer applyer(this);
	utils::for_items(categories, applyer);
	switch (applyer.pco) {
	case Item::PCO_NO:
		break;
	case Item::PCO_LOGOFF:
		requestLogoff();
		break;
	case Item::PCO_REBOOT:
		requestReboot();
		break;
	default:
		assert (false);
	}
}

void TuneApp::Applyer::operator()(Item* item) {
	try {
		item->apply();
		if (item->isChanged()) {
			pco = max(pco, item->getPostChangedOperation());
		}
	} catch (Exception& e) {
		tuneapp->showNoPrivileges(item, e);
	}
}

void TuneApp::showNoPrivileges(const Item* item, const Exception& e) {
	wstring message
			= L"\"" + item->getName() + L"\" ̐ݒ𔽉f邱Ƃo܂łB\n"
			+ L"ݒ𔽉f錠܂B\n"
			+ L"Item ID: " + utils::getIDPath(item) + L"\n"
			+ L"Message: " + e.getMessage();
	::MessageBox(
			hwnd,
			message.c_str(),
			APP_NAME.c_str(),
			MB_OK | MB_ICONWARNING | MB_APPLMODAL);
}

void TuneApp::requestLogoff() {
	exit(	getToolkit()->getStringResource(IDS_REQUEST_LOGOFF),
			EWX_LOGOFF);
}

void TuneApp::requestReboot() {
	try {
		adjustShutdownPrivilige();
		exit(	getToolkit()->getStringResource(IDS_REQUEST_REBOOT),
				EWX_REBOOT);
	} catch (SecurityException) {
		::MessageBox(hwnd,
				getToolkit()->getStringResource(IDS_REQUEST_REBOOT_SKIPED).c_str(), APP_NAME.c_str(),
				MB_OK | MB_ICONWARNING | MB_APPLMODAL);
	}
}

void TuneApp::exit(const wstring& message, int mode) {
	int result = ::MessageBox(
			hwnd,
			message.c_str(), APP_NAME.c_str(),
			MB_YESNO | MB_DEFBUTTON2 | MB_ICONQUESTION | MB_APPLMODAL);
	if (result == IDYES) {
		try {
			exitWindows(mode);
		} catch (SecurityException) {
			::MessageBox(
					hwnd,
					getToolkit()->getStringResource(IDS_SHUTDOWN_SECURITY_EXCEPTION).c_str(),
					APP_NAME.c_str(),
					MB_OK | MB_ICONWARNING | MB_APPLMODAL);
		}
	}
}

void TuneApp::cancel() {
	dispose();
	::PostQuitMessage(0);
}

void TuneApp::save() {
	wstring baseName = showSaveDialog();
	if (baseName == NULL_STRING) return;
	Storage storage(baseName);
	Saver saver(this, &storage);
	utils::for_items(categories, saver);
	try {
		storage.save();
	} catch (Exception) {
		// new ThreadExceptionDialog(e).showDialog();
		throw;
	}
}

void TuneApp::Saver::operator()(Item* item) {
	if (isEnabledAffect(item->getAffectTarget(), tuneapp->affect)) {
		try {
			item->store(storage);
		} catch (Exception) {
			// new ThreadExceptionDialog(e).showDialog();
			throw;
		}
	}
}

wstring TuneApp::showSaveDialog() {
	return showFileDialog(
			getToolkit()->getStringResource(IDS_SAVE),
			awt::FileDialog::SAVE,
			L"tuneapp.reg", L"reg");
}

wstring TuneApp::showFileDialog(
			const wstring& title, awt::FileDialog::Type mode,
			const wstring& name, const wstring& ext) {
	awt::FileDialog dialog(this, title, mode);
	// dialog.setDirectory(new File(".").getAbsolutePath());
	dialog.setFile(name);
	dialog.setExtention(ext);
	dialog.setVisible(true);
	return dialog.getFile();
}

void TuneApp::about() {
	setHelpText(
			APP_NAME + L" version " + VERSION + L" (" + VERSION_SUFIX + L")\r\n" +
			L"Copyright(c) 2002-2005, Win2000Pro `[ibvpX and ~3.EXE,\r\n" +
			L"ALL RIGHTS RESERVED."
	);
}

void TuneApp::playMp3() {
	wstring mp3File = showFileDialog(
			getToolkit()->getStringResource(IDS_PLAY_MP3),
			awt::FileDialog::LOAD,
			L"", L"mp3");
	if (mp3File == NULL_STRING) return;
	UINT result = reinterpret_cast<UINT>(::ShellExecute(
			hwnd,
			L"open",
			mp3File.c_str(), NULL, utils::getFileBasePath(mp3File).c_str(),
			SW_SHOW) );
	if (result <= 32) {
		::MessageBox(
			hwnd,
			getToolkit()->getStringResource(IDS_MP3_FAILED).c_str(),
			getToolkit()->getStringResource(IDS_PLAY_MP3).c_str(),
			MB_OK | MB_ICONWARNING | MB_APPLMODAL);
	}
}

void TuneApp::adjustShutdownPrivilige() throw(SecurityException) {
	auto_handle token;
	TOKEN_PRIVILEGES tkp;
	BOOL result = ::OpenProcessToken(
			::GetCurrentProcess(),
			TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY,
			&token);
	if (result == FALSE) {
		throw SecurityException(L"OpenProcessToken() failed: " + toString(::GetLastError()));
	}
	result = ::LookupPrivilegeValue(
			NULL, SE_SHUTDOWN_NAME, &tkp.Privileges[0].Luid);
	if (result == FALSE) {
		throw SecurityException(L"LookupPrivilegeValue() failed: " + toString(::GetLastError()));
	}
	tkp.PrivilegeCount = 1;
	tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
 	::AdjustTokenPrivileges(token, FALSE, &tkp, 0, NULL, NULL);
	if (::GetLastError() != ERROR_SUCCESS) {
		throw SecurityException(L"AdjustTokenPrivileges() failed: " + toString(::GetLastError()));
	}
}

void TuneApp::exitWindows(int mode) throw(SecurityException) {
	BOOL result = ::ExitWindowsEx(mode, 0);
	if (result == FALSE) {
		throw SecurityException(L"ExitWindowsEx() failed: " + toString(::GetLastError()));
	}
}
