// tuneapp


#include "VersionDependedDocumentBuilder.h"
#include "xml/Text.h"
#include "reg/RegKey.h"
#include <assert.h>


VersionDependedDocumentBuilder::VersionDependedDocumentBuilder(xml::Document* document)
		: document(document), currentNode(document), skipping(false) {
	initVersionInfo();
}

void VersionDependedDocumentBuilder::initVersionInfo() {
	OSVERSIONINFOEX info;
	bool ex = true;
	::ZeroMemory(&info, sizeof(OSVERSIONINFOEX));
	info.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
	BOOL result = ::GetVersionEx(reinterpret_cast<OSVERSIONINFO*>(&info));
	if (result == FALSE) {
		info.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
		result = ::GetVersionEx(reinterpret_cast<OSVERSIONINFO*>(&info));
		if (result = FALSE) {
			throw Exception(L"GetVersionEx() failed");
		}
		ex = false;
	}

	productVersion = (info.dwMajorVersion << 8) | info.dwMinorVersion;
	
	if (ex) {
		productSPVersion = (info.wServicePackMajor << 8) | info.wServicePackMinor;
		if (info.wSuiteMask & VER_SUITE_PERSONAL) {
			productType = PERSONAL;
		} else if (info.wProductType == VER_NT_WORKSTATION) {
			productType = WORKSTATION;
		} else if (info.wProductType == VER_NT_SERVER) {
			productType = SERVER;
		} else {
			throw Exception(L"unknown OS product type: " + toString(info.wProductType));
		}
	} else {
		wstring spString(info.szCSDVersion);
		if (spString.empty()) {
			productSPVersion = 0;
		} else {
			if (!startsWithIgnoreCase(spString, L"Service Pack ") ||
					(spString.length() < 14)) {
				throw Exception(L"unknown OS service pack version: " + spString);
			}
			productSPVersion = (spString.at(13) - '0') << 8;
			if (spString.length() == 15) {
				productSPVersion |= spString.at(14) - 'a' + 1;
			}
		}
		auto_ptr<RegKey> key(new RegKey(
				HKEY_LOCAL_MACHINE,
				L"SYSTEM\\CurrentControlSet\\Control\\ProductOptions", KEY_READ));
		wstring typeString = key->getStringValue(L"ProductType");
		if (!equalsIgnoreCase(typeString, L"WinNT")) {
			productType = WORKSTATION;
		} else if (!equalsIgnoreCase(typeString, L"LanManNT") ||
				!equalsIgnoreCase(typeString, L"ServerNT")) {
			productType = SERVER;
		} else {
			throw Exception(L"unknown OS product type: " + typeString);
		}
	}
}

bool VersionDependedDocumentBuilder::checkVersionDepended(xml::Node* root) throw(ConfigException) {
	const wstring& typeSpec = root->getAttribute(L"productType");
	if (typeSpec != NULL_STRING) {
		pair<Expression, wstring> typeSpecPair = parseSpec(typeSpec);
		ProductType type = getProductType(typeSpecPair.second);
		if (!operate(productType, typeSpecPair.first, type)) {
			return false;
		}
	}
	const wstring& versionSpec = root->getAttribute(L"productVersion");
	if (versionSpec != NULL_STRING) {
		pair<Expression, wstring> versionSpecPair = parseSpec(versionSpec);
		int version = getProductVersion(versionSpecPair.second);
		if (!operate(productVersion, versionSpecPair.first, version)) {
			return false;
		}
	}
	const wstring& spSpec = root->getAttribute(L"productSPVersion");
	if (spSpec != NULL_STRING) {
		pair<Expression, wstring> spSpecPair = parseSpec(spSpec);
		int version = getProductSPVersion(spSpecPair.second);
		if (!operate(productSPVersion, spSpecPair.first, version)) {
			return false;
		}
	}
	return true;
}

pair<VersionDependedDocumentBuilder::Expression, wstring> VersionDependedDocumentBuilder::parseSpec(const wstring& spec) throw(ConfigException) {
	wstring::size_type space = spec.find(' ');
	if (space == wstring::npos) {
		return make_pair(EQUAL, spec);
	} else {
		wstring expressionString = spec.substr(0, space);
		Expression expression;
		if (expressionString == L"==") {
			expression = EQUAL;
		} else if (expressionString == L"<") {
			expression = LESS;
		} else if (expressionString == L">=") {
			expression = GREAT_EQUAL;
		} else {
			throw ConfigException(L"invalid expression: " + spec);
		}
		return make_pair(expression, spec.substr(space + 1));
	}
}

VersionDependedDocumentBuilder::ProductType VersionDependedDocumentBuilder::getProductType(const wstring& typeString) throw(ConfigException) {
	if (equalsIgnoreCase(typeString, L"Personal")) {
		return PERSONAL;
	} else if (equalsIgnoreCase(typeString, L"Workstation")) {
		return WORKSTATION;
	} else if (equalsIgnoreCase(typeString, L"Server")) {
		return SERVER;
	} else {
		throw ConfigException(L"unknown productType: " + typeString);
	}
}

int VersionDependedDocumentBuilder::getProductVersion(const wstring& versionString) throw(ConfigException) {
	try {
		wstring::size_type dot = versionString.find('.');
		if (dot == wstring::npos) {
			// "5" -> 0x500
			return (parseInt(versionString) << 8);
		} else {
			// "5.1" -> 0x501
			return (parseInt(versionString.substr(0, dot)) << 8) |
					parseInt(versionString.substr(dot + 1));
		}
	} catch (NumberFormatException) {
		throw ConfigException(L"invalid productVersion: " + versionString);
	}
}

int VersionDependedDocumentBuilder::getProductSPVersion(const wstring& versionString) throw(ConfigException) {
	try {
		return (parseInt(versionString) << 8);
	} catch (NumberFormatException) {
		throw ConfigException(L"invalid productSPVersion: " + versionString);
	}
}

bool VersionDependedDocumentBuilder::operate(int value1, Expression expression, int value2) {
	switch (expression) {
	case EQUAL:
		return (value1 == value2);
	case LESS:
		return (value1 < value2);
	case GREAT_EQUAL:
		return (value1 >= value2);
	default:
		assert (false);
		return false;
	}
}

void VersionDependedDocumentBuilder::startElement(const wstring& name, const Attributes& attrs) {
	if (skipping) {
		skippingStack.push(name);
		return;
	}

	xml::Element* element = new xml::Element(name);
	element->setAttributes(attrs);
	if (checkVersionDepended(element)) {
		currentNode->add(element);
		currentNode = element;
	} else {
		skipping = true;
		skippingStack.push(name);
	}
}

void VersionDependedDocumentBuilder::endElement(const wstring& name) {
	if (skipping) {
		wstring currentName = skippingStack.top();
		if (currentName != name) {
			throw Exception(L"end name missmatch: " + currentName + L" and /" + name);
		}
		skippingStack.pop();
		if (skippingStack.empty()) {
			skipping = false;
		}
		return;
	}

	if (currentNode->getName() != name) {
		throw Exception(L"end name missmatch: " + currentNode->getName() + L" and /" + name);
	}
	currentNode = currentNode->getParent();
}

void VersionDependedDocumentBuilder::text(const wstring& text) {
	if (skipping) {
		return;
	}

	xml::Text* textNode = new xml::Text(text);
	currentNode->add(textNode);
}

void VersionDependedDocumentBuilder::endDocument() {
	if (currentNode != document) {
		throw Exception(L"no document end");
	}
}
