// AWT less.


#include "Choice.h"

#include <assert.h>


namespace awt {


Choice::Choice() : selectedIndex(-1) {
}

LPCTSTR Choice::getWindowClass() {
	return L"COMBOBOX";
}

DWORD Choice::getWindowStyle() {
	return WS_TABSTOP |
			CBS_DROPDOWNLIST |
			Component::getWindowStyle();
}

HWND Choice::createWindow() {
	HWND hwnd = Component::createWindow();
	for (	vector<wstring>::const_iterator labels = labelList.begin();
			labels != labelList.end();
			++labels) {
		sendAddMessage(hwnd, *labels);
	}
	sendSelectMessage(hwnd);
	return hwnd;
}

void Choice::sendAddMessage(HWND hwnd, const wstring& label) {
	LRESULT index = ::SendMessage(
			hwnd, CB_ADDSTRING, NULL, reinterpret_cast<LPARAM>(label.c_str()));
	if (index < 0) {
		throw Exception(L"SendMessage(CB_ADDSTRING) failed: " + toString(index));
	}
}

void Choice::sendSelectMessage(HWND hwnd) {
	::SendMessage(hwnd, CB_SETCURSEL, selectedIndex, 0);
}

LRESULT Choice::wmCommand(WORD notifyMessage, WORD identifier, bool& consume) {
	switch (notifyMessage) {
	case CBN_SELCHANGE:
		updateSelectedIndex();
		fireItemStateChanged();
		consume = true;
		return 0;
	}
	return Component::wmCommand(notifyMessage, identifier, consume);
}

void Choice::updateSelectedIndex() {
	LRESULT result = ::SendMessage(hwnd, CB_GETCURSEL, 0, 0);
	if (result == CB_ERR) {
		selectedIndex = -1;
	} else {
		selectedIndex = result;
	}
}

void Choice::add(const wstring& label) {
	labelList.push_back(label);
	preferedSizeCache.width = -1;
	if (hwnd) {
		sendAddMessage(hwnd, label);
	}
}

void Choice::removeAll() {
	labelList.clear();
	preferedSizeCache.width = -1;
	if (hwnd) {
		::SendMessage(hwnd, CB_RESETCONTENT, 0, 0);
	}
	selectedIndex = -1;
}

int Choice::getSelectedIndex() const {
	return selectedIndex;
}

void Choice::select(int index) {
	if (	(index != -1) &&
			((index < 0) || (index >= labelList.size())) ) {
		throw IllegalArgumentException(L"index: " + toString(index));
	}
	selectedIndex = index;
	if (hwnd) {	
		sendSelectMessage(hwnd);
	}
}

void Choice::setSize(int width, int height) {
	Component::setSize(width, height);
	integrateSize();
}

void Choice::setBounds(int x, int y, int width, int height) {
	Component::setBounds(x, y, width, height);
	integrateSize();
}

Dimension Choice::getPreferredSize() {
	if (preferedSizeCache.width == -1) {
		assert (hwnd != NULL);
		SIZE size;
		TEXTMETRIC metrics;
		int width = 0;
		auto_hdc hdc(hwnd);
		selectFont(hdc);
		::GetTextMetrics(hdc, &metrics);
		for (	vector<wstring>::const_iterator labels = labelList.begin();
				labels != labelList.end();
				++labels) {
			::GetTextExtentPoint32(hdc, (*labels).c_str(), (*labels).length(), &size);
			width = max(width, size.cx);
		}
		int borderW = ::GetSystemMetrics(SM_CXEDGE) + ::GetSystemMetrics(SM_CXBORDER) * 2;
		int borderH = ::GetSystemMetrics(SM_CYEDGE) + ::GetSystemMetrics(SM_CYBORDER) * 2;
		int buttonW = ::GetSystemMetrics(SM_CXVSCROLL);
		int buttonH = ::GetSystemMetrics(SM_CYVSCROLL);
		preferedSizeCache.width = width + buttonW + borderH * 2;
		preferedSizeCache.height = max(metrics.tmHeight, buttonH) + borderH * 2;
	}
	return preferedSizeCache;
}

/**
 *	integrate height to dropdownlist visible.
 *	call this method post resized.
 */
void Choice::integrateSize() {
	if (hwnd == NULL) return;
	TEXTMETRIC metrics;
	auto_hdc hdc(hwnd);
	selectFont(hdc);
	::GetTextMetrics(hdc, &metrics);
	int borderH = ::GetSystemMetrics(SM_CYBORDER);
	// int height = this->height + metrics.tmHeight * max(labelList.size(), 1) + borderH * 2;
	int height = this->height + ::SendMessage(hwnd, CB_GETITEMHEIGHT, 0, 0) * labelList.size() + borderH * 2;
	// ::MoveWindow(hwnd, x, y, width, height, false);
	::SetWindowPos(
			hwnd, NULL,
			x, y, width, height,
			SWP_NOZORDER | SWP_NOMOVE | SWP_NOREDRAW);
}


} // namespace
