// AWT less.


#include "Component.h"
#include "Window.h"
#include "Toolkit.h"
#include "BorderedPanel.h"
#include "event/MouseListener.h"
#include "event/FocusListener.h"
#include "event/KeyListener.h"
#include "event/ActionListener.h"

#include <sstream>
using namespace std;

#include <windowsx.h>
#include <commctrl.h>

#include <typeinfo.h>
#include <assert.h>


namespace awt {


Component::Component()
		: x(0), y(0), width(0), height(0),
		visible(true), enabled(true),
		foreground(-1), background(-1),
		parent(NULL),
		hwnd(NULL),
		oldWindowProc(NULL),
		backgroundBrush(NULL),
		toolkitCache(NULL), insetsCache(), preferedSizeCache(-1, -1),
		mouseOwner(false) {
}

Component::~Component() {
	if (parent) {
		parent->remove(this);
	}
}


// window creation.


void Component::initCommonControlEx(DWORD dwICC) {
	INITCOMMONCONTROLSEX info;
	::ZeroMemory(&info, sizeof(INITCOMMONCONTROLSEX));
	info.dwSize = sizeof(INITCOMMONCONTROLSEX);
	info.dwICC = dwICC;
	BOOL result = ::InitCommonControlsEx(&info);
	if (result == FALSE) {
		throw Exception(L"InitCommonControlsEx() failed");
	}
}

LPCTSTR Component::getWindowText() {
	return NULL;
}

DWORD Component::getWindowStyle() {
	return WS_CLIPSIBLINGS | WS_CHILD |
			(visible ? WS_VISIBLE : 0) | (enabled ? 0 : WS_DISABLED);
}

DWORD Component::getWindowStyleEx() {
	return WS_EX_NOPARENTNOTIFY;
}

HWND Component::createWindow() {
	HWND hwnd = ::CreateWindowEx(
			getWindowStyleEx(),
			getWindowClass(),
			getWindowText(),
			getWindowStyle(),
			x, y, width, height,
			(parent == NULL ? NULL : parent->getHwnd()),
			NULL,
			getToolkit()->getHinstance(),
			NULL);
	if (hwnd == NULL) {
		throw Exception(L"CreateWindowEx() failed");
	}
	HGDIOBJ font = ::GetStockObject(DEFAULT_GUI_FONT);
	::SendMessage(hwnd, WM_SETFONT, reinterpret_cast<WPARAM>(font), NULL);
	return hwnd;
}


// message processing.


LRESULT Component::windowProc(UINT uMsg, WPARAM wParam, LPARAM lParam) {
	LRESULT result = 0;
	bool consume = false;
	switch (uMsg) {
	case WM_MOUSEMOVE:
		result = wmMouseMove(wParam, lParam, consume);
		break;
	case WM_LBUTTONDOWN:
		result = wmMouseButton(MouseEvent::PRESSED, MouseEvent::BUTTON_L, wParam, lParam, consume);
		break;
	case WM_SETFOCUS:
		fireFocusGained(consume);
		break;
	}
	if (consume) {
		return result;
	}
	return defWindowProc(uMsg, wParam, lParam);
}

LRESULT Component::defWindowProc(UINT uMsg, WPARAM wParam, LPARAM lParam) {
	if (oldWindowProc != NULL) {
		return ::CallWindowProc(oldWindowProc, hwnd, uMsg, wParam, lParam);
	} else {
		return ::DefWindowProc(hwnd, uMsg, wParam, lParam);
	}
}

LRESULT Component::wmCommand(WORD notifyMessage, WORD identifier, bool& consume) {
	return 0;
}

LRESULT Component::wmNotify(WPARAM wParam, LPARAM lParam, bool& consume) {
	return 0;
}


// mouse message


LRESULT Component::wmMouseButton(
		MouseEvent::Type type, MouseEvent::Button button,
		WPARAM wParam, LPARAM lParam,
		bool& consume) {
	int x = GET_X_LPARAM(lParam);
	int y = GET_Y_LPARAM(lParam);
	if (!mouseOwner) {
		processMouseOwned(x, y);
	}
	fireMouseEvent(type, button, x, y, consume);
	return 0;
}

LRESULT Component::wmMouseMove(WPARAM wParam, LPARAM lParam, bool& consume) {
	int x = GET_X_LPARAM(lParam);
	int y = GET_Y_LPARAM(lParam);
	if (!mouseOwner) {
		processMouseOwned(x, y);
	}
	// fire mouse motion event
	return 0;
}

void Component::processMouseOwned(int x, int y) {
	mouseOwner = true;
	getRootWindow()->notifyMouseOwned(this, x, y);
	bool consume = false;
	fireMouseEvent(MouseEvent::ENTERED, MouseEvent::BUTTON_NO, x, y, consume);
}

void Component::processMouseLost(int x, int y) {
	mouseOwner = false;
	bool consume = false;
	fireMouseEvent(MouseEvent::EXITED, MouseEvent::BUTTON_NO, x, y, consume);
}


// key message


LRESULT Component::wmKey(KeyEvent::Type type, WPARAM wParam, LPARAM lParam, bool& consume) {
	fireKeyEvent(type, wParam, consume);
	return 0;
}


// component


Container* Component::getParent() const {
	return parent;
}

Window* Component::getRootWindow() {
	if (Window* window = dynamic_cast<Window*>(this)) {
		return window;
	}
	Container* parent = this->getParent();
	while (true) {
		if (parent == NULL) {
			return NULL;
		} else if (Window* window = dynamic_cast<Window*>(parent)) {
			return window;
		} else {
			parent = parent ->getParent();
		}
	}
}

Point Component::getLocation() const {
	return Point(x, y);
}

void Component::setLocation(int x, int y) {
	if ((this->x == x) && (this->y == y)) return;
	int oldX = this->x;
	int oldY = this->y;
	this->x = x;
	this->y = y;
	if (hwnd) {
		// ::MoveWindow(hwnd, x, y, width, height, FALSE);
		::SetWindowPos(
				hwnd, NULL,
				x, y, width, height,
				SWP_NOZORDER | SWP_NOSIZE);
	}
}

Dimension Component::getSize() const {
	return Dimension(width, height);
}

void Component::setSize(int width, int height) {
	if ((this->width == width) && (this->height == height)) return;
	int oldW = this->width;
	int oldH = this->height;
	this->width = width;
	this->height = height;
	if (hwnd) {
		// ::MoveWindow(hwnd, x, y, width, height, FALSE);
		::SetWindowPos(
				hwnd, NULL,
				x, y, width, height,
				SWP_NOZORDER | SWP_NOMOVE);
		relayout();
	}
}

void Component::setBounds(int x, int y, int width, int height) {
	bool locationChange = ((this->x != x) || (this->y != y));
	bool sizeChange = ((this->width != width) || (this->height != height));
	if (!locationChange && ! sizeChange) return;
	int oldX = this->x;
	int oldY = this->y;
	int oldW = this->width;
	int oldH = this->height;
	this->x = x;
	this->y = y;
	this->width = width;
	this->height = height;
	if (hwnd) {
		// ::MoveWindow(hwnd, x, y, width, height, FALSE);
		::SetWindowPos(
				hwnd, NULL,
				x, y, width, height,
				SWP_NOZORDER);
		if (sizeChange) {
			relayout();
		}
	}
}

Dimension Component::getClientSize() const {
	if (hwnd) {
		RECT rect;
		::GetClientRect(hwnd, &rect);
		return Dimension(rect.right, rect.bottom);
	} else {
		return getSize();
	}
}

Dimension Component::getClientDistance() const {
	if (hwnd) {
		RECT size;
		::GetWindowRect(hwnd, &size);
		RECT client;
		::GetClientRect(hwnd, &client);
		return Dimension(
				size.right - size.left - client.right,
				size.bottom - size.top - client.bottom);
	} else {
		return Dimension();
	}
}

Insets Component::getInsets() const {
	return insetsCache;
}

void Component::setInsets(const Insets& insets) {
	this->insetsCache = insets;
	preferedSizeCache.width = -1;
}

void Component::relayout() {
	getToolkit()->scheduleLayout(this);
}

void Component::relayoutRoot() {
	Window* root = getRootWindow();
	if (root) {
		root->relayout();
	}
}

void Component::doLayout() {
}

bool Component::isVisible() const {
	return visible;
}

void Component::setVisible(bool visible) {
	if (this->visible == visible) return;
	this->visible = visible;
	if (hwnd) {
		if (visible) {
			::ShowWindow(hwnd, SW_SHOW);
		} else {
			::ShowWindow(hwnd, SW_HIDE);
		}
		if (parent) {
			parent->relayout();
		}
		relayoutRoot();
	}
}

bool Component::isEnabled() const {
	return enabled;
}

void Component::setEnabled(bool enabled) {
	if (this->enabled == enabled) return;
	this->enabled = enabled;
	if (hwnd) {
		::EnableWindow(hwnd, enabled);
	}
}

bool Component::isShowing() const {
	if (hwnd == NULL) return false;
	if (!::IsWindowVisible(hwnd)) return false;
	POINT point = { 0, 0 };
	::ClientToScreen(hwnd, &point);
	return (::WindowFromPoint(point) == hwnd);
}


// color.


void Component::setForeground(DWORD foreground) {
	this->foreground = foreground;
	repaint();
}

void Component::setBackground(DWORD background) {
	this->background = background;
	if (backgroundBrush != NULL) {
		::DeleteObject(backgroundBrush);
		backgroundBrush = NULL;
	}
	repaint();
}

HBRUSH Component::wmCtlColor(UINT uMsg, HDC hdc, bool& consume) {
	if (background != -1) {
		if (foreground != -1) {
			::SetTextColor(hdc, foreground); // ignore error
		}
		::SetBkColor(hdc, background);
		if (backgroundBrush == NULL) {
			backgroundBrush = ::CreateSolidBrush(background); // ignore error
		}
		consume = true;
		return backgroundBrush;
	} else if (foreground != -1) {
		HBRUSH stockBrush = reinterpret_cast<HBRUSH>(parent->defWindowProc(
				uMsg, reinterpret_cast<WPARAM>(hdc), reinterpret_cast<LPARAM>(hwnd)));
		::SetTextColor(hdc, foreground); // ignore error
		consume = true;
		return stockBrush;
	} else {
		return NULL;
	}
}


// font.


Dimension Component::getTextExtentPoint(const wstring& text) const {
	assert (hwnd != NULL);
	auto_hdc hdc(hwnd);
	selectFont(hdc);
	SIZE size;
	::GetTextExtentPoint32(hdc, text.c_str(), text.length(), &size);
	return Dimension(size.cx, size.cy);
}

void Component::getTextMetrics(
		const wstring& text, SIZE* size, TEXTMETRIC* metrics) const {
	assert (hwnd != NULL);
	auto_hdc hdc(hwnd);
	selectFont(hdc);
	if (size != NULL) {
		::GetTextExtentPoint32(hdc, text.c_str(), text.length(), size);
	}
	if (metrics != NULL) {
		::GetTextMetrics(hdc, metrics);
	}
}

void Component::selectFont(HDC hdc) const {
	HGDIOBJ font = (HGDIOBJ)::SendMessage(hwnd, WM_GETFONT, 0, 0);
	if (font != NULL) {
		::SelectObject(hdc, font);
	}
}


// focus


bool Component::isFocusable() {
	// Not Implemented;
	if (hwnd) {
		return ((::GetWindowLong(hwnd, GWL_STYLE) & WS_TABSTOP) != 0);
	} else {
		return false;
	}
}

void Component::requestFocus() {
	if (hwnd) {
		::SetFocus(hwnd);
	}
}

static Component* findNextFocusableChildren(Component* component, bool forward) {
	if (Container* container = dynamic_cast<Container*>(component)) {
		int start;
		int end;
		if (forward) {
			start = 0;
			end = container->getComponentCount();
		} else {
			start = container->getComponentCount() - 1;
			end = -1;
		}
		for (int i = start; i != end; i += (forward ? 1 : -1)) {
			Component* focusable
					= findNextFocusableChildren(container->getComponent(i), forward);
			if (focusable) return focusable;
		}
		return NULL;
	} else {
		if (	component->isFocusable() &&
				component->isShowing() && component->isEnabled()) {
			return component;
		} else {
			return NULL;
		}
	}
}

static Component* findNextFocusable(Component* component, bool forward) {
	Container* parent = component->getParent();
	Component* current = component;
	while (true) {
		if (parent == NULL) return NULL;
		Container::const_iterator found
				= find(parent->begin(), parent->end(), current);
		assert (found != parent->end());
		int start = found - parent->begin();
		int end;
		if (forward) {
			start++;
			end = parent->getComponentCount();
		} else {
			start--;
			end = -1;
		}
		for (int i = start; i != end; i += (forward ? 1 : -1)) {
			Component* focusable
					= findNextFocusableChildren(parent->getComponent(i), forward);
			if (focusable) {
				focusable->requestFocus();
				return focusable;
			}
		}
		current = parent;
		parent = parent->getParent();
	}
}

Component* Component::focusNext() {
	return findNextFocusable(this, true);
}

Component* Component::focusPrevious() {
	return findNextFocusable(this, false);
}


// change style


void Component::addWindowStyle(DWORD style) {
	assert (hwnd != NULL);
	SetWindowLong(hwnd, GWL_STYLE,
			::GetWindowLong(hwnd, GWL_STYLE) | style);
}

void Component::removeWindowStyle(DWORD style) {
	assert (hwnd != NULL);
	SetWindowLong(hwnd, GWL_STYLE,
			::GetWindowLong(hwnd, GWL_STYLE) & ~style);
}

bool Component::hasWindowStyle(DWORD style) {
	return ((::GetWindowLong(hwnd, GWL_STYLE) & style) == style);
}


// peer.


Toolkit* Component::getToolkit() const {
	if (parent == NULL) {
		return Toolkit::getDefaultToolkit();
	} else {
		if (toolkitCache == NULL) {
			toolkitCache = parent->getToolkit();
		}
		return toolkitCache;
	}
}

void Component::addNotify() {
	if (!hwnd) {
		hwnd = createWindow();
		getToolkit()->mapHwnd(hwnd, this);
		getToolkit()->subclass(this);
		// repaint();
	}
}

void Component::removeNotify() {
	if (hwnd) {
		getToolkit()->unmapHwnd(hwnd);
		::DestroyWindow(hwnd);
		hwnd = NULL;
		if (backgroundBrush) {
			::DeleteObject(backgroundBrush);
			backgroundBrush = NULL;
		}
		toolkitCache = NULL;
	}
}

HWND Component::getHwnd() const {
	return hwnd;
}

void Component::repaint() {
	repaint(0, 0, width, height);
}

void Component::repaint(int x, int y, int width, int height) {
	repaintThisOnly(x, y, width, height);
}

void Component::repaintThisOnly() {
	repaintThisOnly(0, 0, width, height);
}

void Component::repaintThisOnly(int x, int y, int width, int height) {
	if (!hwnd || !visible) return;
	::RedrawWindow(hwnd, NULL, NULL, RDW_FRAME | RDW_INVALIDATE);
	RECT rect = { x, y, x + width, y + height };
	::InvalidateRect(hwnd, &rect, TRUE);
}


// timer


static void CALLBACK TimerProc(HWND hwnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime) {
	Component* component = Toolkit::getComponentByHwnd(hwnd);
	if (component == NULL) return;
	ActionListener* listener = reinterpret_cast<ActionListener*>(idEvent);
	ActionEvent event(component);
	listener->actionPerformed(event);
}

void Component::setTimer(UINT msec, ActionListener* listener) {
	assert (hwnd != NULL);
	::SetTimer(hwnd, reinterpret_cast<UINT>(listener), msec, TimerProc);
}

void Component::killTimer(ActionListener* listener) {
	::KillTimer(hwnd, reinterpret_cast<UINT>(listener));
}


// listener


void Component::addMouseListener(MouseListener* listener) {
	mouseListeners.add(listener);
}

void Component::removeMouseListener(MouseListener* listener) {
	mouseListeners.remove(listener);
}

void Component::addKeyListener(KeyListener* listener) {
	keyListeners.add(listener);
}

void Component::removeKeyListener(KeyListener* listener) {
	keyListeners.remove(listener);
}

void Component::addFocusListener(FocusListener* listener) {
	focusListeners.add(listener);
}

void Component::removeFocusListener(FocusListener* listener) {
	focusListeners.remove(listener);
}


// event dispatch


void Component::fireMouseEvent(MouseEvent::Type type, MouseEvent::Button button, int x, int y, bool& consume) {
	MouseEvent event(this, type, button, x, y);
	switch (type) {
	case MouseEvent::ENTERED:
		FIRE_EVENT(mouseListeners, MouseListener, mouseEntered, MouseEvent, event);
		break;
	case MouseEvent::EXITED:
		FIRE_EVENT(mouseListeners, MouseListener, mouseExited, MouseEvent, event);
		break;
	case MouseEvent::PRESSED:
		FIRE_EVENT(mouseListeners, MouseListener, mousePressed, MouseEvent, event);
		break;
/*
	case MouseEvent::RELEASED:
		FIRE_EVENT(mouseListeners, MouseListener, mousePressed, MouseEvent, event);
		break;
	case MouseEvent::CLICKED:
		FIRE_EVENT(mouseListeners, MouseListener, mousePressed, MouseEvent, event);
		break;
*/
	default:
		assert (false);
	}
	consume = event.isConsumed();
}

void Component::fireKeyEvent(KeyEvent::Type type, int keyCode, bool& consume) {
	KeyEvent event(this, type, keyCode);
	switch (type) {
	case KeyEvent::PRESSED:
		FIRE_EVENT(keyListeners, KeyListener, keyPressed, KeyEvent, event);
		processFocusRotation(event);
		break;
	case KeyEvent::RELEASED:
		FIRE_EVENT(keyListeners, KeyListener, keyReleased, KeyEvent, event);
		break;
	case KeyEvent::TYPED:
		FIRE_EVENT(keyListeners, KeyListener, keyTyped, KeyEvent, event);
		break;
	default:
		assert (false);
	}
	consume = event.isConsumed();
}

void Component::processFocusRotation(KeyEvent& event) {
	if (event.getKeyCode() == VK_TAB) {
		if (event.getModifiers() == 0) {
			focusNext();
			event.consume();
		} else if (event.getModifiers() == KeyEvent::SHIFT_MASK) {
			focusPrevious();
			event.consume();
		}
	}
}

void Component::fireFocusGained(bool& consume) {
	FocusEvent event(this);
	FIRE_EVENT(focusListeners, FocusListener, focusGained, FocusEvent, event);
	consume = event.isConsumed();
}


#ifdef _DEBUG

void Component::debug(int indent) {
	::OutputDebugString(wstring(indent, ' ').c_str());
	const type_info& info = typeid(*this);
	::OutputDebugStringA(info.name());
	wostringstream buffer;
	buffer	<< L"["
			<< L"this=" << this
			<< L", parent=" << parent
			<< L", x=" << x << L", y=" << y
			<< L", width=" << width << L", height=" << height
			<< L"]" << endl;
	::OutputDebugString(buffer.str().c_str());
}

#endif // _DEBUG


} // namespace
