// AWT less.


#include "Container.h"
#include "NullLayout.h"
#include "Toolkit.h"
#include "event/ContainerListener.h"

#include "../utils.h"

#include <algorithm>
using namespace std;

#include <assert.h>


namespace awt {


Container::Container() : layout(NullLayout::sharedInstance) {
}

Container::~Container() {
}

DWORD Container::getWindowStyle() {
	return WS_GROUP |
			WS_CLIPCHILDREN |
			Component::getWindowStyle();
}

LRESULT Container::windowProc(UINT uMsg, WPARAM wParam, LPARAM lParam) {
	LRESULT result = 0;
	bool consume = false;
	switch (uMsg) {
	case WM_CTLCOLORBTN: // from control to parent
	// case WM_CTLCOLORDLG:
	case WM_CTLCOLOREDIT:
	case WM_CTLCOLORLISTBOX:
	case WM_CTLCOLORSCROLLBAR:
	case WM_CTLCOLORSTATIC:
		{
			HWND targetHwnd = reinterpret_cast<HWND>(lParam);
			Component* component = getToolkit()->getComponentByHwnd(targetHwnd);
			if (component) {
				result = reinterpret_cast<LRESULT>(
						component->wmCtlColor(
						uMsg, reinterpret_cast<HDC>(wParam), consume) );
			}
		}
		break;
	case WM_COMMAND: // from control to parent
		{
			HWND targetHwnd = reinterpret_cast<HWND>(lParam);
			Component* component = getToolkit()->getComponentByHwnd(targetHwnd);
			if (component) {
				result = component->wmCommand(HIWORD(wParam), LOWORD(wParam), consume);
			}
		}
		break;
	case WM_NOTIFY: // from common control to parent
		{
			NMHDR* notifyInfo = reinterpret_cast<NMHDR*>(lParam);
			Component* component = getToolkit()->getComponentByHwnd(notifyInfo->hwndFrom);
			if (component) {
				result = component->wmNotify(notifyInfo->code, lParam, consume);
			}
		}
		break;
	}
	if (consume) {
		return result;
	}
	return Component::windowProc(uMsg, wParam, lParam);
}


// container


void Container::add(Component* child, const LayoutManager::Constraints& constraints) {
	assert (child != NULL);
	assert (child->parent == NULL);
	assert (find(begin(), end(), child) == end());
	childList.push_back(child);
	child->parent = this;
	layout->addLayoutComponent(child, constraints);
	if (hwnd) {
		child->addNotify();
	}
	fireContainerEvent(ContainerEvent::ADD, child);
}

void Container::remove(Component* child) {
	assert (child != NULL);
	iterator found = find(childList.begin(), childList.end(), child);
	assert (found != childList.end());
	child->removeNotify();
	childList.erase(found);
	child->parent = NULL;
	layout->removeLayoutComponent(child);
	fireContainerEvent(ContainerEvent::REMOVE, child);
}

Component* Container::getComponent(int index) const {
	return childList.at(index);
}

int Container::getComponentCount() const {
	return childList.size();
}

const Container::ComponentList& Container::getComponentList() const {
	return childList;
}

Container::const_iterator Container::begin() const {
	return childList.begin();
}

Container::const_iterator Container::end() const {
	return childList.end();
}


// layout


LayoutManager* Container::getLayout() const {
	return layout;
}

void Container::setLayout(LayoutManager* layout) {
	assert (layout != NULL);
	this->layout = layout;
	relayout();
}

Dimension Container::getPreferredSize() {
	return layout->getPreferredLayoutSize(this);
}

void Container::doLayout() {
	layout->layoutContainer(this);
	Component::doLayout();
}


// component


void Container::requestFocus() {
	if (isFocusable()) {
		Component::requestFocus();
	} else {
		Component* focusable = findFocusableChildren(this);
		if (focusable != NULL) {
			focusable->requestFocus();
		}
	}
}

Component* Container::findFocusableChildren(Component* component) {
	if (Container* container = dynamic_cast<Container*>(component)) {
		for (int i = 0; i < container->getComponentCount(); i++) {
			Component* child = container->getComponent(i);
			if (child->isFocusable() && child->isEnabled()) {
				return child;
			}
			Component* focusable = findFocusableChildren(child);
			if (focusable != NULL) return focusable;
		}
		return NULL;
	} else {
		return NULL;
	}
}


// peer


void Container::addNotify() {
	Component::addNotify();
	for_each(begin(), end(), utils::mem_vfun(&Component::addNotify));
}

void Container::removeNotify() {
	for_each(begin(), end(), utils::mem_vfun(&Component::removeNotify));
	Component::removeNotify();
}


// drawing


void Container::repaint(int x, int y, int width, int height) {
	if (!hwnd || !visible) return;
	Component::repaint(x, y, width, height);
	RECT repaintRect = { x, y, x + width, y + height };
	for (int i = 0; i < childList.size(); i++) {
		Component* child = childList[i];
		Point location = child->getLocation();
		Dimension size = child->getSize();
		RECT childRect = { location.x, location.y, location.x + size.width, location.y + size.height };
		RECT dest;
		BOOL result = ::IntersectRect(&dest, &childRect, &repaintRect);
		if (result) {
			child->repaint(
					dest.left - location.x, dest.top - location.y,
					dest.right - dest.left, dest.bottom - dest.top);
		}
	}
}


// listener


void Container::addContainerListener(ContainerListener* listener) {
	containerListners.add(listener);
}

void Container::removeContainerListener(ContainerListener* listener) {
	containerListners.remove(listener);
}

void Container::fireContainerEvent(ContainerEvent::Type type, Component* child) {
	ContainerEvent event(this, type, child);
	switch (type) {
	case ContainerEvent::ADD:
		FIRE_EVENT(containerListners, ContainerListener, componentAdded, ContainerEvent, event);
		break;
	case ContainerEvent::REMOVE:
		FIRE_EVENT(containerListners, ContainerListener, componentRemoved, ContainerEvent, event);
	}
}


#ifdef _DEBUG

void Container::debug(int indent) {
	Component::debug(indent);
	for (	iterator iterator = childList.begin();
			iterator != childList.end();
			++iterator) {
		(*iterator)->debug(indent + 2);
	}
}

#endif // _DEBUG


} // namespace
