// AWT less.


#include "GridFlowLayout.h"

#include <assert.h>


namespace awt {


GridFlowLayout::GridFlowLayout(int rows, int cols)
		: rows(rows), cols(cols),
		halign(FIT), valign(FIT), cellHalign(FIT), cellValign(FIT),
		hgap(0), vgap(0) {
}

void GridFlowLayout::setValign(Alignment valign) {
	this->valign = valign;
}

void GridFlowLayout::setHalign(Alignment halign) {
	this->halign = halign;
}

void GridFlowLayout::setCellValign(Alignment cellValign) {
	this->cellValign = cellValign;
}

void GridFlowLayout::setCellHalign(Alignment cellHalign) {
	this->cellHalign = cellHalign;
}

void GridFlowLayout::setHgap(int hgap) {
	this->hgap = hgap;
}

void GridFlowLayout::setVgap(int vgap) {
	this->vgap = vgap;
}

void GridFlowLayout::addLayoutComponent(Component* child, const Constraints& constraints) {
}

void GridFlowLayout::removeLayoutComponent(Component* child) {
}

Dimension GridFlowLayout::getPreferredLayoutSize(Container* parent) {
	int rows = getRowsCount(parent);
	int cols = getColsCount(parent);

	vector<int> cellWidth;
	vector<int> cellHeight;
	cellWidth.resize(cols);
	cellHeight.resize(rows);
	getCellsSize(&cellWidth, &cellHeight, parent);

	int fullHeight = getFullHeight(cellHeight, parent);
	int fullWidth = getFullWidth(cellWidth, parent);

	Dimension distance = parent->getClientDistance();
	return Dimension(fullWidth + distance.width, fullHeight + distance.height);
}

void GridFlowLayout::layoutContainer(Container* parent) {
	int rows = getRowsCount(parent);
	int cols = getColsCount(parent);

	vector<int> cellWidth;
	vector<int> cellHeight;
	cellWidth.resize(cols);
	cellHeight.resize(rows);
	getCellsSize(&cellWidth, &cellHeight, parent);

	int fullHeight = getFullHeight(cellHeight, parent);
	int fullWidth = getFullWidth(cellWidth, parent);

	Dimension parentSize = parent->getClientSize();
	Insets insets = parent->getInsets();
	int startX;
	int startY;
	switch(halign) {
	case LEFT:
		startX = insets.left;
		break;
	case CENTER:
		startX = insets.left + (parentSize.width - fullWidth) / 2;
		break;
	case RIGHT:
		startX = parentSize.width - fullWidth + insets.left;
		break;
	case FIT:
		startX = insets.left;
		fitLength(&cellWidth, hgap, parentSize.width - insets.left - insets.right);
		break;
	default:
		assert (false);
	}
	switch(valign) {
	case TOP:
		startY = insets.top;
		break;
	case CENTER:
		startY = insets.top + (parentSize.height - fullHeight) / 2;
		break;
	case BOTTOM:
		startY = parentSize.height - fullHeight + insets.top;
		break;
	case FIT:
		startY = insets.top;
		fitLength(&cellHeight, vgap, parentSize.height - insets.top - insets.bottom);
		break;
	default:
		assert (false);
	}

	int countComponent = parent->getComponentCount();
	for (	int i = 0, posY = startY, y = 0;
			y < rows;
			posY += cellHeight.at(y), y++) {
		if ((cellHeight.at(y) > 0) && (posY > startY)) {
			posY += vgap;
		}
		for (	int posX = startX, x = 0;
				(x < cols) && (i < countComponent);
				posX += cellWidth.at(x), x++, i++) {
			Component* child = parent->getComponent(i);
			if (!child->isVisible()) continue;
			if ((cellWidth.at(x) > 0) && (posX > startX)) {
				posX += hgap;
			}
			int cx;
			Dimension size(child->getPreferredSize());
			switch (cellHalign) {
			case LEFT:
				cx = posX;
				break;
			case CENTER:
				cx = posX + (cellWidth.at(x) - size.width) / 2;
				break;
			case RIGHT:
				cx = posX + (cellWidth.at(x) - size.width);
				break;
			case FIT:
				cx = posX;
				size.width = cellWidth.at(x);
				break;
			default:
				assert (false);
			}
			int cy;
			switch (cellValign) {
			case TOP:
				cy = posY;
				break;
			case CENTER:
				cy = posY + (cellHeight.at(y) - size.height) / 2;
				break;
			case BOTTOM:
				cy = posY + (cellHeight.at(y) - size.height);
				break;
			case FIT:
				cy = posY;
				size.height = cellHeight.at(y);
				break;
			default:
				assert (false);
			}
			child->setBounds(cx, cy, size.width, size.height);
		}
	}
}

int GridFlowLayout::getRowsCount(Container* target) {
	if (rows == 0) {
		return (target->getComponentCount() + cols - 1) / cols;
	} else {
		return rows;
	}
}

int GridFlowLayout::getColsCount(Container* target) {
	if (cols == 0) {
		return (target->getComponentCount() + rows - 1) / rows;
	} else {
		return cols;
	}
}

void GridFlowLayout::getCellsSize(
		vector<int>* cellWidth, vector<int>* cellHeight,
		Container* parent) {
	fill(cellWidth->begin(), cellWidth->end(), 0);
	fill(cellHeight->begin(), cellHeight->end(), 0);
	int countComponent = parent->getComponentCount();
	for (int i = 0, y = 0; y < cellHeight->size(); y++) {
		for (int x = 0; (x < cellWidth->size()) && (i < countComponent); x++, i++) {
			Component* child = parent->getComponent(i);
			if (child->isVisible()) {
				Dimension size(child->getPreferredSize());
				cellWidth->at(x) = max(cellWidth->at(x), size.width);
				cellHeight->at(y) = max(cellHeight->at(y), size.height);
			}
		}
	}
}

int GridFlowLayout::getFullHeight(const vector<int>& cellHeight, Container* parent) {
	int height = 0;
	for (int y = 0; y < cellHeight.size(); y++) {
		if (cellHeight.at(y) > 0) {
			if (height > 0) {
				height += vgap;
			}
			height += cellHeight.at(y);
		}
	}
	Insets insets(parent->getInsets());
	height += insets.top + insets.bottom;
	return height;
}

int GridFlowLayout::getFullWidth(const vector<int>& cellWidth, Container* parent) {
	int width = 0;
	for (int x = 0; x < cellWidth.size(); x++) {
		if (cellWidth.at(x) > 0) {
			if (width > 0) {
				width += hgap;
			}
			width += cellWidth.at(x);
		}
	}
	Insets insets(parent->getInsets());
	width += insets.left + insets.right;
	return width;
}

void GridFlowLayout::fitLength(vector<int>* length, int gap, int target) {
	int all = 0;
	for(int i = 0; i < length->size(); i++) {
		if (length->at(i) > 0) {
			if (all > 0) {
				target -= gap;
			}
			all += length->at(i);
		}
	}
	if ((all == 0) || (all == target)) return;
	for (i = 0; i < length->size(); i++) {
		length->at(i) = length->at(i) * target / all;
	}

	all = 0;
	for (i = 0; i < length->size(); i++) {
		all += length->at(i);
	}
	for (i = 0; i < target - all; i++) {
		++(length->at(i));
	}
}


} // namespace
