// AWT less.


#include "List.h"

#include <assert.h>


namespace awt {


List::List() : selectedIndex(-1) {
}

LPCTSTR List::getWindowClass() {
	return L"LISTBOX";
}

DWORD List::getWindowStyle() {
	return WS_TABSTOP | WS_BORDER |
			LBS_NOINTEGRALHEIGHT |
			LBS_NOTIFY |
			Component::getWindowStyle();
}

DWORD List::getWindowStyleEx() {
	return WS_EX_CLIENTEDGE |
			Component::getWindowStyleEx();
}

HWND List::createWindow() {
	HWND hwnd = Component::createWindow();
	::SendMessage(hwnd, LB_INITSTORAGE, labelList.size(), 0); // ignore error
	for	(	vector<wstring>::const_iterator labels = labelList.begin();
			labels != labelList.end();
			++labels) {
		LRESULT index = ::SendMessage(
				hwnd, LB_ADDSTRING, NULL, (LPARAM)(*labels).c_str());
		if (index < 0) {
			throw Exception(L"SendMessage(LB_ADDSTRING) failed: " + toString(index));
		}
	}
	sendSelectMessage(hwnd);
	return hwnd;
}

void List::sendSelectMessage(HWND hwnd) {
	::SendMessage(hwnd, LB_SETCURSEL, selectedIndex, 0);
}

LRESULT List::wmCommand(WORD notifyMessage, WORD identifier, bool& consume) {
	switch (notifyMessage) {
	case LBN_SELCHANGE:
		updateSelectedIndex();
		fireItemStateChanged();
		consume = true;
		return 0;
	}
	return Component::wmCommand(notifyMessage, identifier, consume);
}

void List::updateSelectedIndex() {
	LRESULT result = ::SendMessage(hwnd, LB_GETCURSEL, 0, 0);
	if (result == LB_ERR) {
		selectedIndex = -1;
	} else {
		selectedIndex = result;
	}
}

void List::add(const wstring& label) {
	labelList.push_back(label);
	preferedSizeCache.width = -1;
	if (hwnd) {
		LRESULT index = ::SendMessage(
				hwnd, LB_ADDSTRING, 0, (LPARAM)label.c_str());
		if (index < 0) {
			throw Exception(L"SendMessage(LB_ADDSTRING) failed: " + toString(index));
		}
	}
}

void List::removeAll() {
	labelList.clear();
	preferedSizeCache.width = -1;
	if (hwnd) {
		::SendMessage(hwnd, LB_RESETCONTENT, 0, 0);
	}
	selectedIndex = -1;
}

int List::getSelectedIndex() const {
	return selectedIndex;
}

void List::select(int index) {
	if (	(index != -1) &&
			((index < 0) || (index >= labelList.size())) ) {
		throw IllegalArgumentException(L"index: " + toString(index));
	}
	selectedIndex = index;
	if (hwnd) {
		sendSelectMessage(hwnd);
	}
}

Dimension List::getPreferredSize() {
	if (preferedSizeCache.width == -1) {
		int width = 0;
		int height = 0;
		SIZE size;
		auto_hdc hdc(hwnd);
		selectFont(hdc);
		for (	vector<wstring>::const_iterator labels = labelList.begin();
				labels != labelList.end();
				++labels) {
			::GetTextExtentPoint32(hdc, (*labels).c_str(), (*labels).length(), &size);
			width = max(width, size.cx);
		}
		int itemH = ::SendMessage(hwnd, LB_GETITEMHEIGHT, 0, 0);
		int scrollW = ::GetSystemMetrics(SM_CXVSCROLL);
		int borderW = ::GetSystemMetrics(SM_CXEDGE) + ::GetSystemMetrics(SM_CXBORDER);
		int borderH = ::GetSystemMetrics(SM_CYEDGE) + ::GetSystemMetrics(SM_CYBORDER);
		preferedSizeCache.width = width + scrollW + borderW * 2;
		preferedSizeCache.height = itemH * max(labelList.size(), 1) + borderH * 2;
	}
	return preferedSizeCache;
}


} // namespace
