// AWT less.


#include "NumberField.h"
#include "BorderLayout.h"

#include <assert.h>
#include <limits.h>


namespace awt {


NumberField::NumberField(int value)
		: TextField(toString(value), 7),
		value(value), minimum(INT_MIN), maximum(INT_MAX),
		unitIncStep(1), blockIncStep(10),
		handler(NULL) {
	handler = new Handler(this);
	addKeyListener(handler);
	addTextListener(handler);
}

NumberField::~NumberField() {
	delete handler;
}

DWORD NumberField::getWindowStyle() {
	return ES_NUMBER | ES_RIGHT | ES_NOIME |
			TextComponent::getWindowStyle();
}

HWND NumberField::createWindow() {
	HWND hwnd = TextField::createWindow();
	::ImmAssociateContext(hwnd, NULL);
	return hwnd;
}

int NumberField::getMinimum() {
	return minimum;
}

void NumberField::setMinimum(int minimum) {
	this->minimum = minimum;
}

int NumberField::getMaximum() {
	return maximum;
}

void NumberField::setMaximum(int maximum) {
	this->maximum = maximum;
}

int NumberField::getUnitIncrement() {
	return unitIncStep;
}

void NumberField::setUnitIncrement(int unitIncStep) {
	if (unitIncStep == 0) unitIncStep = 1;
	this->unitIncStep = unitIncStep;
}

int NumberField::getBlockIncrement() {
	return blockIncStep;
}

void NumberField::setBlockIncrement(int blockIncStep) {
	if (blockIncStep == 0) blockIncStep = 1;
	this->blockIncStep = blockIncStep;
}

int NumberField::getValue() {
	return value;
}

void NumberField::setValue(int value, AdjustmentEvent::Type adjustmentType) {
	if (this->value != value) {
		this->value = value;
		updateText();
		fireAdjustmentValueChanged(adjustmentType, value);
	}
}

void NumberField::unitInc() {
	setValue(fixValue(value + unitIncStep), AdjustmentEvent::UNIT_INCREMENT);
}

void NumberField::unitDec() {
	setValue(fixValue(value - unitIncStep), AdjustmentEvent::UNIT_DECREMENT);
}

void NumberField::blockInc() {
	setValue(fixValue(value + blockIncStep), AdjustmentEvent::BLOCK_INCREMENT);
}

void NumberField::blockDec() {
	setValue(fixValue(value - blockIncStep), AdjustmentEvent::BLOCK_DECREMENT);
}

int NumberField::fixValue(int value) {
	value -= value % unitIncStep;
	value = min(maximum, value);
	value = max(minimum, value);
	return value;
}

void NumberField::updateText() {
	wstring text = toString(value);
	if (text != getText()) {
		setText(text);
		setCaretPosition(text.length());
	}
}

void NumberField::updateValue() {
	int value;
	try {
		value = parseInt(getText());
	} catch (NumberFormatException) {
		return;
	}
	value = fixValue(value);
	if (this->value != value) {
		this->value = value;
		fireAdjustmentValueChanged(AdjustmentEvent.TRACK, value);
	}
}

Dimension NumberField::getPreferredSize() {
	if (preferedSizeCache.width == -1) {
		assert (hwnd != NULL);
		auto_hdc hdc(hwnd);
		selectFont(hdc);
		SIZE size;
		::GetTextExtentPoint(hdc, L"8", 1, &size);
		int borderW = ::GetSystemMetrics(SM_CXEDGE);
		int borderH = ::GetSystemMetrics(SM_CYEDGE);
		int gapW = 2;
		int gapH = 1;
		preferedSizeCache.width = size.cx * cols + borderW * 2 + gapW * 2;
		preferedSizeCache.height = size.cy + borderH * 2 + gapH * 2;
	}
	return preferedSizeCache;
}


// KeyListener


NumberField::Handler::Handler(NumberField* target)
		: target(target), recovering(false) {
}

void NumberField::Handler::keyPressed(KeyEvent& event) {
	int modifires = event.getModifiers();
	switch (event.getKeyCode()) {
	case VK_UP:
		if (modifires & KeyEvent.CTRL_MASK) {
			target->setValue(target->getMaximum());
		} else if (modifires & KeyEvent.SHIFT_MASK) {
			target->blockInc();
		} else {
			target->unitInc();
		}
		event.consume();
		break;
	case VK_DOWN:
		if (modifires & KeyEvent.CTRL_MASK) {
			target->setValue(target->getMinimum());
		} else if (modifires & KeyEvent.SHIFT_MASK) {
			target->blockDec();
		} else {
			target->unitDec();
		}
		event.consume();
		break;
	}
}

void NumberField::Handler::keyReleased(KeyEvent& event) {
}

void NumberField::Handler::keyTyped(KeyEvent& event) {
}

void NumberField::Handler::textValueChanged(TextEvent& event) {
	target->updateValue();
	if (recovering) {
		target->killTimer(this);
	}
	target->setTimer(2000, this);
	recovering = true;
}

void NumberField::Handler::actionPerformed(ActionEvent& event) {
	target->updateText();
}


} // namespace
