// AWT less.


#include "Toolkit.h"
#include "../utils.h"


#include <assert.h>


namespace awt {


Toolkit* Toolkit::defaultToolkit = new Toolkit();

HINSTANCE Toolkit::hInstance = NULL;

ATOM Toolkit::frameClass = 0;


void Toolkit::init(
		HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow) {
	Toolkit::hInstance = hInstance;
}

Toolkit* Toolkit::getDefaultToolkit() {
	return defaultToolkit;
}


Toolkit::Toolkit() {
}


// message processing


int Toolkit::messageLoop() {
	BOOL result;
	MSG msg;
	while ((result = ::GetMessage(&msg, NULL, 0, 0)) != 0) {
		if (result == -1) {
			throw Exception(L"GetMessage() failed");
		}

		bool consume = false;
		switch (msg.message) {
		case WM_KEYDOWN:
			wmKey(msg.hwnd, KeyEvent::PRESSED, msg.wParam, msg.lParam, consume);
			break;
		case WM_KEYUP:
			wmKey(msg.hwnd, KeyEvent::RELEASED, msg.wParam, msg.lParam, consume);
			break;
		case WM_CHAR:
			wmKey(msg.hwnd, KeyEvent::TYPED, msg.wParam, msg.lParam, consume);
			break;
		}

		if (!consume) {
			::TranslateMessage(&msg);
			::DispatchMessage(&msg);
		}

		processDelayedLayout();
	}
	return msg.wParam;
}

LRESULT Toolkit::wmKey(HWND hwnd, KeyEvent::Type type, WPARAM wParam, LPARAM lParam, bool& consume) {
	Component* component = getComponentByHwnd(hwnd);
	fireKeyEvent(component, type, wParam, consume);
	if (consume) return 0;
	if (component) {
		return component->wmKey(type, wParam, lParam, consume);
	} else {
		return 0;
	}
}

LRESULT WINAPI Toolkit::WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	assert (hwnd != NULL);
	Component* component = getComponentByHwnd(hwnd);
	if (component == NULL) {
		return ::DefWindowProc(hwnd, uMsg, wParam, lParam);
	} else {
		return component->windowProc(uMsg, wParam, lParam);
	}
}


// global key listener


void Toolkit::addKeyListener(KeyListener* listener) {
	keyListeners.add(listener);
}

void Toolkit::removeKeyListener(KeyListener* listener) {
	keyListeners.remove(listener);
}

void Toolkit::fireKeyEvent(Component* component, KeyEvent::Type type, int keyCode, bool& consume) {
	KeyEvent event(component, type, keyCode);
	switch (type) {
	case KeyEvent::PRESSED:
		FIRE_EVENT(keyListeners, KeyListener, keyPressed, KeyEvent, event);
		break;
	case KeyEvent::RELEASED:
		FIRE_EVENT(keyListeners, KeyListener, keyReleased, KeyEvent, event);
		break;
	case KeyEvent::TYPED:
		FIRE_EVENT(keyListeners, KeyListener, keyTyped, KeyEvent, event);
		break;
	default:
		assert (false);
	}
	consume = event.isConsumed();
}


// delayed layout


void Toolkit::scheduleLayout(Component* component) {
	if (find(layoutQueue.begin(), layoutQueue.end(), component) == layoutQueue.end()) {
		layoutQueue.push_back(component);
	}
}

void Toolkit::processDelayedLayout() {
	while (layoutQueue.size() > 0) {
		Component* component = layoutQueue.front();
		layoutQueue.pop_front();
		component->doLayout();
		if (HIWORD(::GetQueueStatus(QS_ALLEVENTS)) != 0) break;
	}
}


// toolkit


wstring Toolkit::getStringResource(UINT id) {
	// Not Implemented
	// extend length
	auto_array<wchar_t> buffer(256);
	int length = ::LoadString(hInstance, id, buffer, 256);
	return wstring(buffer, length);
}

HINSTANCE Toolkit::getHinstance() {
	return hInstance;
}

ATOM Toolkit::getFrameClass() {
	if (frameClass == 0) {
		frameClass = registerClass(L"awt::Frame", CS_PARENTDC);
	}
	return frameClass;
}

ATOM Toolkit::registerClass(const wstring& className, UINT style) {
	WNDCLASSEX info;
	::ZeroMemory(&info, sizeof(WNDCLASSEX));
	info.cbSize = sizeof(WNDCLASSEX);
 	info.style = style;
	info.lpfnWndProc = ::DefWindowProc;
	info.hInstance = hInstance;
	info.hCursor = ::LoadCursor(NULL, MAKEINTRESOURCE(IDC_ARROW));
	info.hbrBackground = (HBRUSH)COLOR_WINDOW;
	info.lpszClassName = className.c_str();
	ATOM atom = ::RegisterClassEx(&info);
	if (atom == 0) {
		throw Exception(L"RegisterClassEx(" + className + L") failed");
	}
	return atom;
}

void Toolkit::mapHwnd(HWND hwnd, Component* component) {
	assert (hwnd != NULL);
	LONG_PTR oldPtr = ::SetWindowLongPtr(
			hwnd, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(component));
	if (oldPtr != NULL) {
		throw Exception(L"window user data already set");
	}
}

void Toolkit::unmapHwnd(HWND hwnd) {
	assert (hwnd != NULL);
	LONG_PTR oldPtr = ::SetWindowLongPtr(hwnd, GWLP_USERDATA, 0);
	if (oldPtr == NULL) {
		throw Exception(L"window user data broken");
	}
}

Component* Toolkit::getComponentByHwnd(HWND hwnd) {
	if (hwnd == NULL) return NULL;
	if (::GetWindowLongPtr(hwnd, GWLP_WNDPROC)
			!= reinterpret_cast<LONG_PTR>(WindowProc)) return NULL;
	return reinterpret_cast<Component*>(::GetWindowLongPtr(hwnd, GWLP_USERDATA));
}

void Toolkit::subclass(Component* component) {
	assert (component->getHwnd() != NULL);
	WNDPROC oldWindowProc = reinterpret_cast<WNDPROC>(
			::SetWindowLongPtr(
					component->getHwnd(),
					GWLP_WNDPROC, reinterpret_cast<LONG_PTR>(WindowProc)) );
	if (oldWindowProc != WindowProc) {
		component->oldWindowProc = oldWindowProc;
	}
}


} // namespace
