// AWT less.


#include "Tree.h"

#include <assert.h>


namespace awt {


Tree::Tree() : showRootHandles(false), selectedNode(NULL), eventLockCount(0) {
	rootNode.handle = TVI_ROOT;
}

LPCTSTR Tree::getWindowClass() {
	return WC_TREEVIEW;
}

DWORD Tree::getWindowStyle() {
	return	WS_TABSTOP |
			TVS_HASBUTTONS | TVS_HASLINES |
			(showRootHandles ? TVS_LINESATROOT : 0) |
			TVS_SHOWSELALWAYS | TVS_DISABLEDRAGDROP | TVS_NOTOOLTIPS |
			Component::getWindowStyle();
}

DWORD Tree::getWindowStyleEx() {
	return WS_EX_CLIENTEDGE |
			Component::getWindowStyleEx();
}

HWND Tree::createWindow() {
	initCommonControlEx(ICC_TREEVIEW_CLASSES);
	HWND hwnd = Component::createWindow();
	::SendMessage(hwnd, TVM_SETUNICODEFORMAT, TRUE, 0);
	sendAddChildrenMessage(hwnd, rootNode);
	if (selectedNode != NULL) {
		sendSelectMessage(hwnd, *selectedNode);
	}
	return hwnd;
}

LRESULT Tree::wmNotify(UINT notifyCode, LPARAM lParam, bool& consume) {
	switch (notifyCode) {
	case TVN_GETDISPINFO:
		{
			NMTVDISPINFO* info = reinterpret_cast<NMTVDISPINFO*>(lParam);
			getDispInfo(info);
			consume = true;
		}
		return 0;
	case TVN_SELCHANGING:
		consume = true;
		return 0;
	case TVN_SELCHANGED:
		if (!eventLocked()) {
			NMTREEVIEW* info = reinterpret_cast<NMTREEVIEW*>(lParam);
			updateSelectedNode(info->itemNew);
			fireItemStateChanged();
			consume = true;
			return 0;
		}
		break;
	case TVN_ITEMEXPANDING:
		consume = true;
		return 0;
	case TVN_ITEMEXPANDED:
		{
			NMTREEVIEW* info = reinterpret_cast<NMTREEVIEW*>(lParam);
			updateNodeExpanded(info->itemNew, info->action);
			consume = true;
		}
		return 0;
	}
	return Component::wmNotify(notifyCode, lParam, consume);
}

Tree::Node* Tree::getNode(HTREEITEM handle) const {
	assert (handle != NULL);
	TVITEM info;
	::ZeroMemory(&info, sizeof(TVITEM));
	info.mask = TVIF_HANDLE | TVIF_PARAM;
	info.hItem = handle;
	LRESULT result = ::SendMessage(
			hwnd, TVM_GETITEM, 0, reinterpret_cast<LPARAM>(&info));
	if (result == FALSE) {
		throw Exception(L"SendMessage(TVM_GETITEM) failed");
	}
	return reinterpret_cast<Node*>(info.lParam);
}

void Tree::getDispInfo(NMTVDISPINFO* info) {
	Node* node = reinterpret_cast<Node*>(info->item.lParam);
	if (info->item.mask & TVIF_TEXT) {
		info->item.pszText = const_cast<wchar_t*>(node->label.c_str());
	}
	if (info->item.mask & TVIF_CHILDREN) {
		info->item.cChildren = (node->childList.empty() ? 0 : 1);
	}
}

void Tree::sendAddMessage(HWND hwnd, const Node& parent, Node& node) {
	assert (parent.handle != NULL);
	assert (node.handle == NULL);
	TVINSERTSTRUCT info;
	::ZeroMemory(&info, sizeof(TVINSERTSTRUCT));
	info.hParent = parent.handle;
	info.hInsertAfter = TVI_LAST;
	info.item.mask = TVIF_TEXT | TVIF_PARAM | TVIF_STATE | TVIF_CHILDREN;
	info.item.pszText = LPSTR_TEXTCALLBACK;
	info.item.lParam = reinterpret_cast<LPARAM>(&node);
	info.item.state = (node.expanded ? TVIS_EXPANDED : 0);
	info.item.cChildren = I_CHILDRENCALLBACK;
	node.handle = reinterpret_cast<HTREEITEM>(
			::SendMessage(
					hwnd, TVM_INSERTITEM, 0, reinterpret_cast<LPARAM>(&info)) );
	if (node.handle == NULL) {
		throw Exception(L"SendMessage(TVM_INSERTITEM) failed");
	}

	// muu
	if ((&parent != &rootNode) && parent.isExpanded()) {
		sendExpandMessage(hwnd, parent);
	}
}

void Tree::sendAddChildrenMessage(HWND hwnd, Node& node) {
	for (	NodeList::iterator iterator = node.childList.begin();
			iterator != node.childList.end();
			++iterator) {
		sendAddMessage(hwnd, node, **iterator);
		sendAddChildrenMessage(hwnd, **iterator);
	}
}

void Tree::sendSelectMessage(HWND hwnd, const Node& node) {
	assert (node.handle != NULL);
	LRESULT result = ::SendMessage(
			hwnd, TVM_SELECTITEM, TVGN_CARET, reinterpret_cast<LPARAM>(node.handle));
	if (result == FALSE) {
		throw Exception(L"SendMessage(TVM_SELECTITEM) failed)");
	}
}

void Tree::sendExpandMessage(HWND hwnd, const Node& node) {
	assert (node.handle != NULL);
	LRESULT result = ::SendMessage(
			hwnd, TVM_EXPAND,
			(node.expanded ? TVE_EXPAND : TVE_COLLAPSE),
			reinterpret_cast<LPARAM>(node.handle));
	if (result == FALSE) {
		// throw Exception(L"SendMessage(TVM_EXPAND) failed");
	}
}

void Tree::updateSelectedNode(const TVITEM& newItem) {
	if (newItem.hItem == NULL) {
		selectedNode = NULL;
	} else if (newItem.state & TVIS_SELECTED) {
		selectedNode = reinterpret_cast<Node*>(newItem.lParam);
	} else {
		selectedNode = NULL;
	}
}

void Tree::updateNodeExpanded(const TVITEM& newItem, UINT action) {
	Node* node = reinterpret_cast<Node*>(newItem.lParam);
	switch (action) {
	case TVE_EXPAND:
		node->expanded = true;
		break;
	case TVE_COLLAPSE:
		node->expanded = false;
		break;
	default:
		throw Exception(L"unknown tree expanded action: " + toString(action));
	}
}

Tree::Node* Tree::getRootNode() {
	return &rootNode;
}

Tree::Node* Tree::addNode(const wstring& label, void* value) {
	return addNode(&rootNode, label, value);
}

Tree::Node* Tree::addNode(Node* parent, const wstring label, void* value) {
	assert (parent != NULL);
	Node* node = parent->addNode();
	node->label = label;
	node->value = value;
	if (parent == &rootNode) {
		preferedSizeCache.width = -1;
	}
	if (hwnd) {
		sendAddMessage(hwnd, *parent, *node);
	}
	return node;
}

void Tree::removeAll() {
	rootNode.clearChildren();
	preferedSizeCache.width = -1;
	if (hwnd) {
		lockEvent();
		LRESULT result = ::SendMessage(
				hwnd, TVM_DELETEITEM, 0, reinterpret_cast<LPARAM>(rootNode.handle));
		unlockEvent();
		if (result == FALSE) {
			throw Exception(L"SendMesssage(TVM_DELETEITEM) failed");
		}
	}
	selectedNode = NULL;
}

Tree::Node* Tree::getSelectedNode() {
	return selectedNode;
}

void Tree::select(Node* node) {
	assert (node != NULL);
	selectedNode = node;
	if (hwnd) {
		sendSelectMessage(hwnd, *node);
	}
}

Tree::Node* Tree::getNextNode(Node* node) {
	return getNextNode(node, TVGN_NEXTVISIBLE);
}

Tree::Node* Tree::getPreviousNode(Node* node) {
	return getNextNode(node, TVGN_PREVIOUSVISIBLE);
}

Tree::Node* Tree::getNextNode(Node* node, WPARAM wParam) {
	assert (hwnd != NULL);
	assert (node->handle != NULL);
	HTREEITEM handle = reinterpret_cast<HTREEITEM>(
			::SendMessage(
					hwnd, TVM_GETNEXTITEM,
					wParam, reinterpret_cast<LPARAM>(node->handle)) );
	if (handle == NULL) {
		return NULL;
	} else {
		return getNode(handle);
	}
}

Tree::Node* Tree::findValue(Node* node, void* value) {
	for (	NodeList::iterator iterator = node->childList.begin();
			iterator != node->childList.end();
			++iterator) {
		if ((*iterator)->value == value) {
			return *iterator;
		}
	}
	return NULL;
}

void Tree::setShowsRootHandles(bool showRootHandles) {
	if (this->showRootHandles == showRootHandles) return;
	this->showRootHandles = showRootHandles;
	preferedSizeCache.width = -1;
	if (hwnd) {
		// Not Implemented: no work. why?
		if (showRootHandles) {
			addWindowStyle(TVS_LINESATROOT);
		} else {
			removeWindowStyle(TVS_LINESATROOT);
		}
	}
}

void Tree::setExpanded(Node* node, bool expanded) {
	node->expanded = expanded;
	if (hwnd) {
		sendExpandMessage(hwnd, *node);
	}
}

Dimension Tree::getPreferredSize() {
	if (preferedSizeCache.width == -1) {
		TEXTMETRIC metrics;
		auto_hdc hdc(hwnd);
		selectFont(hdc);
		::GetTextMetrics(hdc, &metrics);
		int handleW = ::SendMessage(hwnd, TVM_GETINDENT, 0, 0);
		int width = getMaxWidth(hdc, rootNode, handleW, (showRootHandles ? 1 : 0));
		int scrollH = ::GetSystemMetrics(SM_CXVSCROLL);
		int borderW = ::GetSystemMetrics(SM_CXEDGE) + ::GetSystemMetrics(SM_CXBORDER);
		int borderH = ::GetSystemMetrics(SM_CYEDGE) + ::GetSystemMetrics(SM_CYBORDER);
		int focusH = ::GetSystemMetrics(SM_CYBORDER);
		preferedSizeCache.width = width + scrollH + borderW * 2;
		preferedSizeCache.height
				= metrics.tmHeight * max(rootNode.childList.size(), 1)
				+ focusH * rootNode.childList.size() * 2 + borderH * 2;
	}
	return preferedSizeCache;
}

int Tree::getMaxWidth(HDC hdc, const Node& parent, int handleW, int indent) {
	int width = 0;
	SIZE size;
	for (	NodeList::const_iterator nodes = parent.childList.begin();
			nodes != parent.childList.end();
			++nodes) {
		const wstring& label = (*nodes)->label;
		::GetTextExtentPoint32(hdc, label.c_str(), label.length(), &size);
		int focusW = ::GetSystemMetrics(SM_CXBORDER);
		width = max(width, handleW * indent + size.cx + focusW * 2);
		width = max(width, getMaxWidth(hdc, **nodes, handleW, indent + 1));
	}
	return width;
}

Tree::Node* Tree::getClosestNode(const Point& location) const {
	assert (hwnd != NULL);
	TVHITTESTINFO info;
	info.pt.x = location.x;
	info.pt.y = location.y;
	::SendMessage(hwnd, TVM_HITTEST, 0, reinterpret_cast<LPARAM>(&info));
	if (info.hItem == NULL) return NULL;
	return getNode(info.hItem);
}


// event lock


bool Tree::eventLocked() {
	return (eventLockCount > 0);
}

void Tree::lockEvent() {
	eventLockCount++;
}

void Tree::unlockEvent() {
	eventLockCount--;
}


// Node


Tree::Node::Node() : handle(NULL), value(NULL), expanded(false) {
}

Tree::Node::~Node() {
	for (	NodeList::const_iterator iterator = childList.begin();
			iterator != childList.end();
			++iterator) {
		delete *iterator;
	}
}

Tree::Node* Tree::Node::addNode() {
	Node* node = new Node();
	childList.push_back(node);
	return node;
}

void Tree::Node::clearChildren() {
	for (	NodeList::const_iterator iterator = childList.begin();
			iterator != childList.end();
			++iterator) {
		delete *iterator;
	}
	childList.clear();
}

void* Tree::Node::getValue() const {
	return value;
}

bool Tree::Node::isExpanded() const {
	return expanded;
}

int Tree::Node::getChildCount() const {
	return childList.size();
}

Tree::Node* Tree::Node::getChild(int index) {
	return childList.at(index);
}


} // namespace
