// AWT less.


#include "Window.h"
#include "Toolkit.h"
#include "event/WindowListener.h"

#include <assert.h>


namespace awt {


Window::Window() : minimumSize(-1, -1), mouseOwner(NULL) {
	visible = false;
	x = CW_USEDEFAULT;
	y = CW_USEDEFAULT;
}

Window::~Window() {
}

DWORD Window::getWindowStyle() {
	return WS_OVERLAPPED |
			(Container::getWindowStyle() & ~WS_CHILD);
}


// message processing


LRESULT Window::windowProc(UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch (uMsg) {
	case WM_MOVE:
		{
			RECT rect;
			::GetWindowRect(hwnd, &rect);
			wmMove(rect.left, rect.top);
		}
		break;
	case WM_SIZING:
		{
			BOOL fullDrag;
			::SystemParametersInfo(SPI_GETDRAGFULLWINDOWS, 0, &fullDrag, 0);
			if (fullDrag) {
				RECT* rect = reinterpret_cast<RECT*>(lParam);
				wmSizing(wParam, rect->right - rect->left, rect->bottom - rect->top);
			}
		}
		break;
	case WM_SIZE:
		{
			RECT rect;
			::GetWindowRect(hwnd, &rect);
			wmSize(wParam, rect.right - rect.left, rect.bottom - rect.top);
		}
		break;
	case WM_CLOSE:
		fireWindowClosing();
		break;
	case WM_SYSCOMMAND:
		wmSysCommand(wParam);
		break;
	case WM_GETMINMAXINFO:
		wmGetMinMaxInfo(reinterpret_cast<MINMAXINFO*>(lParam));
		break;
	}
	return Container::windowProc(uMsg, wParam, lParam);
}

void Window::wmMove(int x, int y) {
	this->x = x;
	this->y = y;
}

void Window::wmSizing(WPARAM wParam, int width, int height) {
	if ((this->width == width) && (this->height == height)) return;
	this->width = width;
	this->height = height;
	relayout();
}

void Window::wmSize(WPARAM wParam, int width, int height) {
	if ((this->width == width) && (this->height == height)) return;
	this->width = width;
	this->height = height;
	relayout();
}

void Window::wmSysCommand(WPARAM wParam) {
	switch (wParam) {
	case SC_CLOSE:
		fireWindowClosing();
		break;
	}
}

void Window::wmGetMinMaxInfo(MINMAXINFO* info) {
	if (minimumSize.width != -1) {
		info->ptMinTrackSize.x = minimumSize.width;
		info->ptMinTrackSize.y = minimumSize.height;
	}
}


// peer


Toolkit* Window::getToolkit() const {
	if (toolkitCache == NULL) {
		toolkitCache = Toolkit::getDefaultToolkit();
	}
	return toolkitCache;
}

void Window::addNotify() {
	Container::addNotify();
	RECT rect;
	::GetWindowRect(hwnd, &rect);
	x = rect.left;
	y = rect.top;
	width = rect.right - rect.left;
	height = rect.bottom - rect.top;
}


// mouse owner


void Window::notifyMouseOwned(Component* mouseOwner, int x, int y) {
	if (this->mouseOwner != NULL) {
		POINT point = { x, y };
		::MapWindowPoints(mouseOwner->getHwnd(), this->mouseOwner->getHwnd(), &point, 1);
		this->mouseOwner->processMouseLost(x, y);
	}
	this->mouseOwner = mouseOwner;
}


// component


void Window::setIcon(HICON smallIcon, HICON largeIcon) {
	// Not Implemented.
	// select size.
	assert (hwnd != NULL);
	::SendMessage(hwnd, WM_SETICON, ICON_SMALL, reinterpret_cast<WPARAM>(smallIcon));
	::SendMessage(hwnd, WM_SETICON, ICON_BIG, reinterpret_cast<WPARAM>(largeIcon));
}

void Window::setMinimumSize(Dimension minimumSize) {
	this->minimumSize = minimumSize;
}

void Window::setVisible(bool visible) {
	if (visible) {
		if (!hwnd) {
			addNotify();
		}
	}
	Container::setVisible(visible);
}

/*
void Window::setVisible(int nCmdShow) {
	// Not Implemented
}
*/

void Window::pack() {
	addNotify();
	setSize(getPreferredSize());
}

void Window::dispose() {
	setVisible(false);
	removeNotify();
}


// listener


void Window::addWindowListener(WindowListener* listener) {
	windowListeners.add(listener);
}

void Window::removeWindowListener(WindowListener* listener) {
	windowListeners.remove(listener);
}

void Window::fireWindowClosing() {
	WindowEvent event(this);
	FIRE_EVENT(windowListeners, WindowListener, windowClosing, WindowEvent, event);
}


} // namespace
