// tuneapp.


#include "ExecItem.h"

#include "../TuneApp.h"
#include "../ConfigReader.h"
#include "../reg/RegKey.h"

#include "../awt/Window.h"
#include "../awt/FlowLayout.h"

#include "../utils.h"

#include <vector>
using namespace std;


void ExecItem::setParams(xml::Node* params) throw(ConfigException) {
	for (	xml::Node::const_iterator children = params->begin();
			children != params->end();
			++children) {
		xml::Node* child = *children;
		if (child->getType() == xml::Node::ELEMENT) {
			const wstring& name = child->getName();
			if (name == L"registry") {
				setRegistryParams(child);
			} else if (name == L"file") {
				setFileParams(child);
			} else if (name == L"directory") {
				setDirectoryParams(child);
			}
		}
	}
}

void ExecItem::setRegistryParams(xml::Node* params) throw(ConfigException) {
	wstring buttonName;
	wstring key;
	wstring value;
	vector<wstring> fileNameList;

	for (	xml::Node::const_iterator children = params->begin();
			children != params->end();
			++children) {
		xml::Node* child = *children;
		if (child->getType() == xml::Node::ELEMENT) {
			const wstring& name = child->getName();
			if (name == L"name") {
				buttonName = child->getText();
			} else if (name == L"key") {
				key = child->getText();
			} else if (name == L"value") {
				value = child->getText();
			} else if (name == L"file") {
				fileNameList.push_back(child->getText());
			}
		}
	}

	wstring path;
	try {
		auto_ptr<RegKey> reg(new RegKey(key, KEY_READ));
		path = reg->getStringValue(value);
	} catch (RegKeyException) {
		// Not Impelmented.
		// Notify to user?
		return;
	}
	for (	vector<wstring>::const_iterator files = fileNameList.begin();
			files != fileNameList.end();
			++files) {
		wstring file = utils::getAbsolutePath(utils::concatPath(path, *files));
		if (::GetFileAttributes(file.c_str()) != -1) {
			addCommand(buttonName, file);
			return;
		}
	}
}

void ExecItem::setFileParams(xml::Node* params) throw(ConfigException) {
	wstring buttonName;
	wstring file;

	for (	xml::Node::const_iterator children = params->begin();
			children != params->end();
			++children) {
		xml::Node* child = *children;
		if (child->getType() == xml::Node::ELEMENT) {
			const wstring& name = child->getName();
			if (name == L"name") {
				buttonName = child->getText();
			} else if (name == L"file") {
				file = child->getText();
			}
		}
	}
	addCommand(buttonName, file);
}

void ExecItem::setDirectoryParams(xml::Node* params) throw(ConfigException) {
	wstring path;
	vector<wstring> wildcardList;

	for (	xml::Node::const_iterator children = params->begin();
			children != params->end();
			++children) {
		xml::Node* child = *children;
		if (child->getType() == xml::Node::ELEMENT) {
			const wstring& name = child->getName();
			if (name == L"path") {
				path = child->getText();
			} else if (name == L"file") {
				wildcardList.push_back(child->getText());
			}
		}
	}

	for (	vector<wstring>::const_iterator iterator = wildcardList.begin();
			iterator != wildcardList.end();
			++iterator) {
		wstring wildcard = utils::concatPath(path, *iterator);
		WIN32_FIND_DATA findInfo;
		auto_findfile handle(::FindFirstFile(wildcard.c_str(), &findInfo));
		if (handle == INVALID_HANDLE_VALUE) {
			// Not Implemented.
			return;
		}
		while (true) {
			wstring file(findInfo.cFileName);
			wstring baseName = utils::getFileBaseName(file);
			wstring::size_type lastDot = baseName.rfind('.');
			if (lastDot == wstring::npos) lastDot = baseName.length();
			wstring name = baseName.substr(0, lastDot);
			addCommand(name, utils::concatPath(path, file));
			BOOL result = ::FindNextFile(handle, &findInfo);
			if (result == FALSE) {
				DWORD error = ::GetLastError();
				if (error == ERROR_NO_MORE_FILES) {
					break;
				} else {
					// Not Implemented.
					break;
				}
			}
		}
	}
}

void ExecItem::addCommand(const wstring& name, const wstring& command) {
	commandMap[name] = command;
}

awt::Component* ExecItem::getViewComponent() {
	return new View(this);
}

bool ExecItem::isChanged() {
	return false;
}

void ExecItem::setDefault() {}
void ExecItem::apply() {}
void ExecItem::cancel() {}
void ExecItem::store(Storage* storage) {}

const wstring& ExecItem::getShortHelpText() const {
	// apps/ ?
	return NULL_STRING;
}


// View.

ExecItem::View::View(ExecItem* target) : target(target) {
	createComponents();
}

void ExecItem::View::createComponents() {
	setLayout(utils::leftFlowLayout);
	for (	map<wstring, wstring>::const_iterator iterator = target->commandMap.begin();
			iterator != target->commandMap.end();
			++iterator) {
		awt::Button* button = createButton((*iterator).first, (*iterator).second);
		add(button);
	}
}

awt::Button* ExecItem::View::createButton(const wstring& label, const wstring& command) {
	awt::Button* button = new awt::Button(label);
	button->addActionListener(new Executioner(this, command));
	return button;
}

void ExecItem::View::Executioner::actionPerformed(awt::ActionEvent& event) {
	target->exec(command);
}

void ExecItem::View::exec(const wstring& command) {
	try {
		utils::exec(
				utils::concatPath(utils::getSystemDirectory(), L"cmd.exe"),
				L"/c start \"\" \"" + command + L"\"");
	} catch (IOException) {
		::MessageBox(
				getRootWindow()->getHwnd(),
				(command + L" Nł܂łB").c_str(),
				TuneApp::APP_NAME.c_str(),
				MB_OK | MB_ICONWARNING | MB_APPLMODAL) ;
	}
}
