// tuneapp.


#include "AbstractRegItem.h"

#include "../ConfigReader.h"
#include "../utils.h"


AbstractRegItem::AbstractRegItem()
		: regKey(NULL_STRING), regValue(NULL_STRING), regType(REG_NONE),
		defaultValue(NULL_STRING), defaultMean(NULL_STRING), deleteForDefault(false) {
}

void AbstractRegItem::setParams(xml::Node* params) throw(ConfigException) {
	for (	xml::Node::const_iterator children = params->begin();
			children != params->end();
			++children) {
		xml::Node* child = *children;
		if (child->getType() == xml::Node::ELEMENT) {
			const wstring& name = child->getName();
			if (name == L"key") {
				regKey = child->getText();
			} else if (name == L"value") {
				regValue = child->getText();
			} else if (name == L"type") {
				regType = getRegType(child->getText());
			} else if (name == L"default") {
				defaultValue = child->getText();
				defaultMean = child->getAttribute(L"mean");
				deleteForDefault = (child->getAttribute(L"delete") != NULL_STRING);
			}
		}
	}
}

const wstring& AbstractRegItem::getRegKeyName() const {
	return regKey;
}

void AbstractRegItem::setRegKeyName(const wstring& regKey) {
	this->regKey = regKey;
}

const wstring& AbstractRegItem::getRegValueName() const {
	return regValue;
}

void AbstractRegItem::setRegValueName(const wstring& regValue) {
	this->regValue = regValue;
}

DWORD AbstractRegItem::getRegType() const {
	return regType;
}

const wstring& AbstractRegItem::getShortHelpText() const {
	if (shortHelp.empty()) {
		shortHelp = utils::concatPath(regKey, regValue);
		// truncate it
	}
	return shortHelp;
}

RegKey* AbstractRegItem::openRegKey(REGSAM access) const throw(RegKeyException) {
	if (regKey == NULL_STRING) {
		throw ConfigException(L"no set key name");
	}
	return new RegKey(regKey, access);
}

const wstring& AbstractRegItem::getDefaultValueString() const {
	return defaultValue;
}

const wstring& AbstractRegItem::getDefaultMeanString() const {
	return defaultMean;
}

bool AbstractRegItem::isDeleteForDefault() const {
	return deleteForDefault;
}

void AbstractRegItem::setDeleteForDefault(bool deleteForDefault) {
	this->deleteForDefault = deleteForDefault;
}

void AbstractRegItem::store(Storage* storage) {
	store(storage->getRegistryStorage());
}


// Utils.

/**
*	Convert registry type string to int value.
*
*	@param typeString &quot;REG_*&quot;
*	@return RegKey.REG_*
*	@exception IllegalArgumentException
*/
DWORD AbstractRegItem::getRegType(const wstring& typeString) throw(IllegalArgumentException) {
	if (typeString == L"REG_SZ") {
		return REG_SZ;
	} else if (typeString == L"REG_EXPAND_SZ") {
		return REG_EXPAND_SZ;
	} else if (typeString == L"REG_MULTI_SZ") {
		return REG_MULTI_SZ;
	} else if (typeString == L"REG_DWORD") {
		return REG_DWORD;
	} else if (typeString == L"REG_BINARY") {
		return REG_BINARY;
	} else {
		throw IllegalArgumentException(L"typeString:" + typeString);
	}
}

/**
*	Convert String to byte[].
*	@param String hex data likes &quot;00,01,02,ff&quot;
*	@param byte[] return { 0x00, 0x01, 0x02, 0xff }
*	@exception NumberFormatException
*/
BYTE* AbstractRegItem::hexStringToBytes(const wstring& string, int* length) throw(NumberFormatException) {
	vector<BYTE> buffer;
	int hex = 0;
	bool stored = false;
	for (int i = 0; i < string.length(); i++) {
		wchar_t c = string[i];
		if (c == ',') {
			buffer.push_back((BYTE)hex);
			hex = 0;
			stored = false;
		} else {
			hex = hex * 16 + utils::hexToInt(c);
			stored = true;
		}
	}
	if (stored) {
		buffer.push_back((BYTE)hex);
	}
	BYTE* bytes = new BYTE[buffer.size()];
	copy(buffer.begin(), buffer.end(), bytes);
	*length = buffer.size();
	return bytes;
}

/**
*	Convert byte[] to String.
*	@param bytes byte[] { 0x00, 0x01, 0x02, 0xff }
*	@return String likes &quot;00,01,02,ff&quot;
*/
void AbstractRegItem::bytesToHexString(const BYTE* bytes, int length, wstring* string) {
	for (int i = 0; i < length; i++) {
		if (i > 0) {
			(*string) += L",";
		}
		(*string) += utils::byteToHexString(bytes[i]);
	}
}
