// tuneapp.


#include "RegDeleteSubKeysItem.h"

#include "../ConfigReader.h"
#include "../lang.h"
#include "../utils.h"


RegDeleteSubKeysItem::RegDeleteSubKeysItem()
		: cantOperation(false), cantExport(false) {
}

void RegDeleteSubKeysItem::setParams(xml::Node* params) throw(ConfigException) {
	for (	xml::Node::const_iterator children = params->begin();
			children != params->end();
			++children) {
		xml::Node* child = *children;
		if (child->getType() == xml::Node::ELEMENT) {
			const wstring& name = child->getName();
			if (name == L"save") {
				setSaveParams(child);
			} else if (name == L"backup") {
				backupName = child->getText();
			}
		}
	}
	AbstractRegCheckItem::setParams(params);

	init();
}

void RegDeleteSubKeysItem::setSaveParams(xml::Node* params) {
	for (	xml::Node::const_iterator children = params->begin();
			children != params->end();
			++children) {
		xml::Node* child = *children;
		if (ConfigReader::isElementName(child, L"key")) {
			addSaveSubKey(child->getText());
		}
	}
}

void RegDeleteSubKeysItem::addSaveSubKey(const wstring& subkey) {
	saveSubKeyList.insert(toLowerCase(subkey));
}

bool RegDeleteSubKeysItem::isSaveSubKey(const wstring& subkey) const {
	return (saveSubKeyList.find(toLowerCase(subkey)) != saveSubKeyList.end());
}

void RegDeleteSubKeysItem::init() {
	readRegistry();
	if (systemValue) {
		if (!checkBackupFile()) {
			cantOperation = true;
		}
	}
	// make short help
}

void RegDeleteSubKeysItem::readRegistry() {
	try {
		auto_ptr<RegKey> key(openRegKey(KEY_READ));
		systemValue = readRegistry(key.get());
	} catch (RegKeyException) {
		systemValue = true;
		cantExport = true;
	}
	currentValue = systemValue;
}

bool RegDeleteSubKeysItem::readRegistry(RegKey* key) throw(RegKeyException) {
	vector<wstring> subkeys;
	key->getSubKeys(&subkeys);
	for (int i = 0; i < subkeys.size(); i++) {
		if (!isSaveSubKey(subkeys[i])) {
			return false;
		}
	}
	return true;
}

bool RegDeleteSubKeysItem::checkBackupFile() const {
	return (::GetFileAttributes(backupName.c_str()) != -1);
}

awt::Component* RegDeleteSubKeysItem::getViewComponent() {
	awt::Component* component = AbstractRegCheckItem::getViewComponent();
	if (cantOperation) {
		component->setEnabled(false);
	}
	return component;
}

void RegDeleteSubKeysItem::apply() throw(RegKeyException, IOException) {
	if (cantOperation) return;
	if (currentValue == systemValue) return;

	writeRegistry();
	setChanged();
}

void RegDeleteSubKeysItem::writeRegistry() throw(RegKeyException, IOException) {
	if (currentValue) {
		backupRegistry();
		deleteRegistry();
	} else {
		importBackupFile();
	}
}

void RegDeleteSubKeysItem::writeRegistry(RegKey* key, bool value) throw(RegKeyException) {
	assert (false);
}

void RegDeleteSubKeysItem::backupRegistry() throw(RegKeyException, IOException) {
	RegistryStorage backup(backupName);
	RegistryStorage::Key* key = backup.getKey(getRegKeyName());
	key->exportCurrentAll();
	backup.save();
}

void RegDeleteSubKeysItem::deleteRegistry() throw(RegKeyException) {
	auto_ptr<RegKey> key(openRegKey(KEY_WRITE | KEY_READ));
	deleteRegistry(key.get());
	systemValue = true;
}

void RegDeleteSubKeysItem::deleteRegistry(RegKey* key) throw(RegKeyException) {
	vector<wstring> subkeys;
	key->getSubKeys(&subkeys);
	for (int i = 0; i < subkeys.size(); i++) {
		if (!isSaveSubKey(subkeys[i])) {
			key->deleteSubKey(subkeys[i]);
		}
	}
}

void RegDeleteSubKeysItem::importBackupFile() throw(IOException) {
	wstring command = utils::concatPath(utils::getSystemDirectory(), L"cmd.exe");
	wstring backupName = utils::getAbsolutePath(this->backupName);
	wstring args = L"/c start \"\" regedit.exe /s \"" + backupName + L"\"";
	STARTUPINFO startInfo;
	::ZeroMemory(&startInfo, sizeof(STARTUPINFO));
	startInfo.cb = sizeof(STARTUPINFO);
	PROCESS_INFORMATION processInfo;
	BOOL result = ::CreateProcess(
			command.c_str(), const_cast<wchar_t*>(args.c_str()),
			NULL, NULL, FALSE, CREATE_DEFAULT_ERROR_MODE | CREATE_NO_WINDOW, NULL, NULL,
			&startInfo, &processInfo);
	if (result == FALSE) {
		throw IOException(L"CreateProcess(" + command + L") failed: " + toString(::GetLastError()));
	}
	auto_handle hProcess(processInfo.hProcess);
	DWORD waitExit = ::WaitForSingleObject(hProcess, INFINITE);
	if (waitExit != WAIT_OBJECT_0) {
		throw IOException(L"Failed to wait for process import registry: " + toString(::GetLastError()));
	}
	DWORD exitCode;
	result = ::GetExitCodeProcess(hProcess, &exitCode);
	if (result == FALSE) {
		throw IOException(L"Can't get result code for import registry: " + toString(::GetLastError()));
	}
	if (exitCode != 0) {
		throw IOException(L"Failed to import registry: " + toString(exitCode));
	}
	systemValue = false;
}

void RegDeleteSubKeysItem::store(RegistryStorage* storage) {
	if (cantExport) return;
	if (!currentValue) return;

	const wstring& keyName = getRegKeyName();
	RegistryStorage::Key* key = storage->getKey(keyName);
	try {
		key->setDeleteAtFirst();
		key->exportCurrentValues();
		for (	set<wstring>::const_iterator subKeys = saveSubKeyList.begin();
				subKeys != saveSubKeyList.end();
				++subKeys) {
			wstring subKey = utils::concatPath(keyName, *subKeys);
			try {
				// check existed
				delete new RegKey(subKey, KEY_READ);
			} catch (RegKeyException) {
				continue;
			}
			key = storage->getKey(subKey);
			key->exportCurrentValues();
			// Request for Enhancement: open close open close... pool keys ?
		}
	} catch (RegKeyException) {
		storage->cancelExport(keyName);
		// Not Implemented
		// new ThreadExceptionDialog(e).showDialog();
		throw;
	}
}
