// tuneapp.


#include "RegKey.h"

#include "../types.h"


static inline const wchar_t* valueNameCstr(const wstring& valueName) {
	if (valueName.empty()) {
		throw IllegalArgumentException(L"empty value name");
	} else if (valueName == NULL_STRING) {
		return NULL;
	} else {
		return valueName.c_str();
	}
}

HKEY RegKey::getRootID(const wstring& rootString) throw(IllegalArgumentException) {
	if (rootString == L"HKEY_CLASSES_ROOT") {
		return HKEY_CLASSES_ROOT;
	} else if (rootString == L"HKEY_LOCAL_MACHINE") {
		return HKEY_LOCAL_MACHINE;
	} else if (rootString == L"HKEY_CURRENT_USER") {
		return HKEY_CURRENT_USER;
	} else if (rootString == L"HKEY_USERS") {
		return HKEY_USERS;
	} else {
		throw IllegalArgumentException(L"rootString: " + rootString);
	}
}

bool RegKey::isStringType(DWORD type) {
	return	(type == REG_SZ) ||
			(type == REG_EXPAND_SZ) ||
			(type == REG_MULTI_SZ);
}

void RegKey::deleteKeyIfExists(const wstring& keyName) throw(RegKeyException) {
	wstring::size_type lastSeparator = keyName.rfind('\\');
	if (lastSeparator == wstring::npos) {
		throw IllegalArgumentException(L"no separator: " + keyName);
	}
	wstring parentKeyName = keyName.substr(0, lastSeparator);
	wstring subKeyName = keyName.substr(lastSeparator + 1);
	if (subKeyName.empty()) {
		throw IllegalArgumentException(keyName);
	}
	try {
		// check parent key existed
		delete new RegKey(parentKeyName, KEY_READ);
	} catch (RegKeyException e) {
		return;
	}
	auto_ptr<RegKey> key(new RegKey(parentKeyName, KEY_WRITE | KEY_READ));
	key->deleteSubKey(subKeyName);
}

void RegKey::deleteValueIfExists(const wstring& keyName, const wstring& valueName)
		throw(RegKeyException) {
	try {
		// check key existed
		delete new RegKey(keyName, KEY_READ);
	} catch (RegKeyException) {
		return;
	}
	auto_ptr<RegKey> key(new RegKey(keyName, KEY_WRITE | KEY_READ));
	key->deleteValue(valueName);
}


RegKey::RegKey(const wstring& keyName, REGSAM access) throw(RegKeyException) {
	wstring::size_type firstSeparator = keyName.find('\\');
	wstring rootString;
	HKEY rootID;
	if (firstSeparator == wstring::npos) {
		rootString = keyName;
	} else {
		rootString = keyName.substr(0, firstSeparator);
	}
	rootID = getRootID(rootString);
	wstring subKey;
	if (firstSeparator == wstring::npos) {
		// subKey = "";
	} else {
		subKey = keyName.substr(firstSeparator + 1);
	}
	createHandle(rootID, subKey, access);
}

RegKey::RegKey(HKEY rootID, const wstring& subKey, REGSAM access)
		throw(RegKeyException) {
	createHandle(rootID, subKey, access);
}

void RegKey::createHandle(HKEY rootID, const wstring& subKey, REGSAM access)
		throw(RegKeyException) {
	int error;
	if (access & KEY_CREATE_SUB_KEY) {
		error = ::RegCreateKeyEx(
				rootID, (subKey.empty() ? NULL : subKey.c_str()),
				0, NULL, 0, access, NULL, &handle, NULL);
	} else {
		error = ::RegOpenKeyEx(
				rootID, (subKey.empty() ? NULL : subKey.c_str()),
				0, access, &handle);
	}
	if (error != ERROR_SUCCESS) {
		throw RegKeyException(L"open key: " + subKey);
	}
}

RegKey::~RegKey() {
	if (::RegCloseKey(handle) != ERROR_SUCCESS) {
		throw bad_exception("close key");
	}
}

void RegKey::flush() throw(RegKeyException) {
	if (::RegFlushKey(handle) != ERROR_SUCCESS) {
		throw RegKeyException(L"flush key");
	}
}

bool RegKey::hasValue(const wstring& valueName) {
	DWORD type;
	int error = ::RegQueryValueEx(
			handle, valueNameCstr(valueName), NULL,
			&type, NULL, NULL);
	return (error == ERROR_SUCCESS);
}

int RegKey::getIntValue(const wstring& valueName) throw(RegKeyException) {
	DWORD type;
	DWORD size;
	int error = ::RegQueryValueEx(
			handle, valueNameCstr(valueName), NULL,
			&type, NULL, &size);
	if (error != ERROR_SUCCESS) {
		throw RegKeyException(L"query value: " + valueName);
	}
	if (type != REG_DWORD) {
		throw RegKeyException(L"type missmatch: " + toString(type));
	}
	if (size != sizeof(int)) {
		throw RegKeyException(L"size missmatch: " + toString(size));
	}
	int value;
	error = ::RegQueryValueEx(
			handle, valueNameCstr(valueName), NULL, NULL,
			reinterpret_cast<LPBYTE>(&value), &size);
	if (error != ERROR_SUCCESS) {
		throw RegKeyException(L"query value: " + valueName);
	}
	return value;
}

int RegKey::getIntValue(const wstring& valueName, int defaultValue) {
	try {
		return getIntValue(valueName);
	} catch (RegKeyException e) {
		return defaultValue;
	}
}

BYTE* RegKey::getBytesValue(const wstring& valueName, int* length) throw(RegKeyException) {
	DWORD type;
	DWORD size;
	int error = ::RegQueryValueEx(
			handle, valueNameCstr(valueName), NULL,
			&type, NULL, &size);
	if (error != ERROR_SUCCESS) {
		throw RegKeyException(L"query value: " + valueName);
	}
	if (type != REG_BINARY) {
		throw RegKeyException(L"type missmatch: " + toString(type));
	}
	BYTE* buffer = new BYTE[size];
	error = ::RegQueryValueEx(
			handle, valueNameCstr(valueName), NULL, NULL,
			buffer, &size);
	if (error != ERROR_SUCCESS) {
		delete[] buffer;
		throw RegKeyException(L"query value: " + valueName);
	}
	*length = size;
	return buffer;
}

wstring RegKey::getStringValue(const wstring& valueName) throw(RegKeyException) {
	DWORD type;
	DWORD size;
	int error = ::RegQueryValueEx(
			handle, valueNameCstr(valueName), NULL,
			&type, NULL, &size);
	if (error != ERROR_SUCCESS) {
		throw RegKeyException(L"query value: " + valueName);
	}
	if (!isStringType(type)) {
		throw RegKeyException(L"type missmatch: " + toString(type));
	}
	auto_array<wchar_t> buffer(size / 2);
	error = ::RegQueryValueEx(
			handle, valueNameCstr(valueName), NULL, NULL,
			reinterpret_cast<LPBYTE>(buffer.ptr()), &size);
	if (error != ERROR_SUCCESS) {
		throw RegKeyException(L"query value: " + valueName);
	}
	return wstring(buffer.ptr(), size / 2 - 1);
}

wstring RegKey::getStringValue(const wstring& valueName, const wstring& defaultValue) throw(RegKeyException) {
	try {
		return getStringValue(valueName);
	} catch (RegKeyException) {
		return defaultValue;
	}
}

void RegKey::setValue(const wstring& valueName, int value) throw(RegKeyException) {
	int error = ::RegSetValueEx(
			handle, valueNameCstr(valueName), NULL,
			REG_DWORD, reinterpret_cast<LPBYTE>(&value), sizeof(value));
	if (error != ERROR_SUCCESS) {
		throw RegKeyException(L"set value: " + valueName);
	}
}

void RegKey::setValue(const wstring& valueName, BYTE* value, int length)
		throw(RegKeyException) {
	int error = ::RegSetValueEx(
			handle, valueNameCstr(valueName), NULL,
			REG_BINARY, reinterpret_cast<LPBYTE>(value), length);
	if (error != ERROR_SUCCESS) {
		throw RegKeyException(L"set value: " + valueName);
	}
}

void RegKey::setValue(const wstring& valueName, DWORD type, const wstring& value)
		throw(RegKeyException) {
	if (!isStringType(type)) {
		throw IllegalArgumentException(L"type: " + toString(type));
	}
	int error = ::RegSetValueEx(
			handle, valueNameCstr(valueName), NULL,
			type, reinterpret_cast<CONST BYTE*>(value.c_str()), (value.length() + 1) * 2);
	if (error != ERROR_SUCCESS) {
		throw RegKeyException(L"set value: " + valueName);
	}
}

void RegKey::deleteSubKey(const wstring& keyName) throw(RegKeyException) {
	if (keyName == NULL_STRING) {
		throw IllegalArgumentException(L"keyName");
	}
	if (::RegDeleteKey(handle, keyName.c_str()) != ERROR_SUCCESS) {
		throw RegKeyException(L"delete sub key: " + keyName);
	}
}

void RegKey::deleteValue(const wstring& valueName) throw(RegKeyException) {
	if (::RegDeleteValue(handle, valueNameCstr(valueName))
			!= ERROR_SUCCESS) {
		throw RegKeyException(L"delete value: " + valueName);
	}
}

vector<wstring>* RegKey::getSubKeys(vector<wstring>* subkeys) throw(RegKeyException) {
	DWORD count;
	DWORD maxNameLength;
	int error = ::RegQueryInfoKey(
			handle, NULL, NULL, NULL,
			&count, &maxNameLength,
			NULL, NULL, NULL, NULL, NULL, NULL);
	if (error != ERROR_SUCCESS) {
		throw RegKeyException(L"query info key");
	}
	subkeys->reserve(count);
	int bufferLength = maxNameLength + 1;
	auto_array<wchar_t> buffer(bufferLength);
	for (int i = 0; i < count; i++) {
		DWORD length = bufferLength;
		error = ::RegEnumKeyEx(
				handle, i,
				buffer, &length,
				NULL, NULL, NULL, NULL);
		if (error != ERROR_SUCCESS) {
			throw RegKeyException(L"enumerate subkey");
		}
		subkeys->push_back(wstring(buffer, length));
	}
	return subkeys;
}

map<wstring, DWORD>* RegKey::getValueTypes(map<wstring, DWORD>* valueTypes)
		throw(RegKeyException) {
	DWORD count;
	DWORD maxNameLength;
	int error = ::RegQueryInfoKey(
			handle, NULL, NULL, NULL, NULL, NULL, NULL,
			&count, &maxNameLength, NULL, NULL, NULL);
	if (error != ERROR_SUCCESS) {
		throw RegKeyException(L"query info key");
	}
	int bufferLength = maxNameLength + 1;
	auto_array<wchar_t> buffer(bufferLength);
	for (int i = 0; i < count; i++) {
		DWORD length = bufferLength;
		DWORD type;
		error = ::RegEnumValue(
				handle, i,
				buffer, &length, NULL,
				&type, NULL, NULL);
		if (error != ERROR_SUCCESS) {
			throw RegKeyException(L"enumerate subkey");
		}
		(*valueTypes)[wstring(buffer, length)] = type;
	}
	return valueTypes;
}
