// tuneapp.


#include "RegToggleItem.h"

#include "../ConfigReader.h"

#include "../ui/ToggleItemView.h"


RegToggleItem::RegToggleItem() : defaultSelection(0) {
}

void RegToggleItem::setParams(xml::Node* params) throw(ConfigException) {
	for (	xml::Node::const_iterator children = params->begin();
			children != params->end();
			++children) {
		xml::Node* child = *children;
		if (child->getType() == xml::Node::ELEMENT) {
			const wstring& name = child->getName();
			if (name == L"select") {
				setSelectParams(child);
			}
		}
	}
	AbstractRegItem::setParams(params);

	init();
}

void RegToggleItem::setSelectParams(xml::Node* params) throw(ConfigException) {
	for (	xml::Node::const_iterator children = params->begin();
			children != params->end();
			++children) {
		xml::Node* child = *children;
		if (ConfigReader::isElementName(child, L"item")) {
			const wstring& value = child->getAttribute(L"value");
			const wstring& label = child->getText();
			valuesString.push_back(value);
			labels.push_back(label);
		}
	}
}

static int searchArray(const vector<wstring>& array, const wstring& pattern) {
	vector<wstring>::const_iterator found
			= find(array.begin(), array.end(), pattern);
	if (found == array.end()) {
		return -1;
	}
	return found - array.begin();
}

void RegToggleItem::init() throw(ConfigException) {
	const wstring& defaultValueString = getDefaultValueString();
	defaultSelection = searchArray(valuesString, defaultValueString);
	if (defaultSelection == -1) {
		throw ConfigException(L"default no match: " + defaultValueString);
	}

	readRegistry();
}

void RegToggleItem::readRegistry() {
	try {
		auto_ptr<RegKey> key(openRegKey(KEY_READ));
		systemSelection = readRegistry(key.get());
	} catch (RegKeyException) {
		systemSelection = defaultSelection;
	}
	currentSelection = systemSelection;
	// make short help
}

int RegToggleItem::readRegistry(RegKey* key) throw(RegKeyException) {
	int systemSelection;
	switch (getRegType()) {
	case REG_DWORD:
		{
			int systemValue = key->getIntValue(
					getRegValueName(), parseInt(getDefaultValueString()) );
			wstring systemValueString = toString(systemValue);
			systemSelection = searchArray(valuesString, systemValueString);
			if (systemSelection == -1) {
				systemSelection = defaultSelection;
			}
			break;
		}
	case REG_BINARY:
		{
			int length;
			auto_array<BYTE> systemValueBytes(key->getBytesValue(getRegValueName(), &length));
			if (systemValueBytes == NULL) {
				systemSelection = defaultSelection;
			} else {
				wstring systemValueString;
				bytesToHexString(systemValueBytes, length, &systemValueString);
				systemSelection = searchArray(valuesString, systemValueString);
				if (systemSelection == -1) {
					systemSelection = defaultSelection;
				}
			}
			break;
		}
	case REG_SZ:
	case REG_EXPAND_SZ:
	case REG_MULTI_SZ:
		{
			wstring systemValueString = key->getStringValue(getRegValueName(), NULL_STRING);
			if (systemValueString == NULL_STRING) {
				systemSelection = defaultSelection;
			} else {
				systemSelection = searchArray(valuesString, systemValueString);
				if (systemSelection == -1) {
					systemSelection = defaultSelection;
				}
			}
			break;
		}
	default:
		throw ConfigException(L"RegType");
	}
	return systemSelection;
}

awt::Component* RegToggleItem::getViewComponent() {
	return new ToggleItemView(this);
}

int RegToggleItem::getSystemSelection() const {
	return systemSelection;
}

int RegToggleItem::getCurrentSelection() const {
	return currentSelection;
}

int RegToggleItem::getDefaultSelection() const {
	return defaultSelection;
}

int RegToggleItem::getCountItems() const {
	return valuesString.size();
}

void RegToggleItem::setCurrentSelection(int currentSelection) {
	if (this->currentSelection == currentSelection) return;
	this->currentSelection = currentSelection;
	firePropertyChange(PROP_CURRENT_VALUE);
}

const wstring& RegToggleItem::getLabel(int index) const {
	return labels[index];
}

void RegToggleItem::setDefault() {
	setCurrentSelection(defaultSelection);
}

void RegToggleItem::apply() throw(RegKeyException) {
	if (currentSelection == systemSelection) return;

	writeRegistry();
	setChanged();
}

void RegToggleItem::writeRegistry() throw(RegKeyException) {
	auto_ptr<RegKey> key(openRegKey(KEY_WRITE | KEY_READ));
	writeRegistry(key.get());
	systemSelection = currentSelection;
}

void RegToggleItem::writeRegistry(RegKey* key) throw(RegKeyException) {
	if ((currentSelection == defaultSelection) && isDeleteForDefault()) {
		key->deleteValue(getRegValueName());
	} else {
		const wstring& valueString = valuesString[currentSelection];
		int type = getRegType();
		switch (type) {
		case REG_DWORD:
			{
				int valueInt = parseInt(valueString);
				key->setValue(getRegValueName(), valueInt);
				break;
			}
		case REG_BINARY:
			{
				int length;
				auto_array<BYTE> valueBytes(hexStringToBytes(valueString, &length));
				key->setValue(getRegValueName(), valueBytes, length);
				break;
			}
		case REG_SZ:
		case REG_EXPAND_SZ:
		case REG_MULTI_SZ:
			key->setValue(getRegValueName(), type, valueString);
			break;
		default:
			throw ConfigException(L"RegType");
		}
	}
}

void RegToggleItem::cancel() {
	setCurrentSelection(systemSelection);
}

void RegToggleItem::store(RegistryStorage* storage) {
	RegistryStorage::Key* key = storage->getKey(getRegKeyName());
	if ((currentSelection == defaultSelection) && isDeleteForDefault()) {
		key->setDeleteValue(getRegValueName());
	} else {
		const wstring& valueString = valuesString[currentSelection];
		int type = getRegType();
		switch (type) {
		case REG_DWORD:
			{
				int valueInt = parseInt(valueString);
				key->setDWORD(getRegValueName(), valueInt);
				break;
			}
		case REG_BINARY:
			{
				int length;
				auto_array<BYTE> valueBytes(hexStringToBytes(valueString, &length));
				key->setBINARY(getRegValueName(), valueBytes, length);
				break;
			}
		case REG_SZ:
		case REG_EXPAND_SZ:
		case REG_MULTI_SZ:
			key->setString(getRegValueName(), type, valueString);
			break;
		default:
			throw ConfigException(L"RegType");
		}
	}
}
