// tuneapp.


#include "RegistryStorage.h"

#include "../types.h"
#include "../utils.h"
#include "../ConfigException.h"
#include "../IOException.h"

#include <sstream>
#include <iomanip>
using namespace std;


RegistryStorage::RegistryStorage(const wstring& baseName) : baseName(baseName) {
}

RegistryStorage::~RegistryStorage() {
	for (	map<wstring, Key*>::iterator iterator = keyMap.begin();
			iterator != keyMap.end();
			++iterator) {
		delete (*iterator).second;
	}
}

RegistryStorage::Key* RegistryStorage::getKey(const wstring& keyName) {
	wstring keyNameLow = toLowerCase(keyName);
	Key* key = utils::find_ptr(keyMap, keyNameLow);
	if (key == NULL) {
		key = new Key(this, keyName);
		keyMap[keyNameLow] = key;
	} else if (dynamic_cast<DeleteKey*>(key)) {
		throw ConfigException(L"conflict set and delete: " + keyName);
	}
	return key;
}

void RegistryStorage::deleteKey(const wstring& keyName) {
	wstring keyNameLow = toLowerCase(keyName);
	Key* key = utils::find_ptr(keyMap, keyNameLow);
	if (key == NULL) {
		key = new DeleteKey(this, keyName);
		keyMap[keyNameLow] = key;
	} else if (dynamic_cast<DeleteKey*>(key) == NULL) {
		throw ConfigException(L"conflict set and delete: " + keyName);
	}
}

void RegistryStorage::cancelExport(const wstring& keyName) {
	wstring keyNameLow = toLowerCase(keyName);
	map<wstring, Key*>::iterator found = keyMap.find(keyNameLow);
	if (found != keyMap.end()) {
		delete (*found).second;
		keyMap.erase(found);
	}
}

void RegistryStorage::save() throw(IOException) {
	// Not Implemented: support reg format version 4 for NT4.0 compatible

	wostringstream out;
	write(out);
	
	wstring fileName = baseName;
	if (!endsWithIgnoreCase(baseName, L".reg")) {
		fileName += L".reg";
	}

	auto_handle file(::CreateFile(
			fileName.c_str(),
			GENERIC_WRITE, 0, NULL, CREATE_ALWAYS,
			FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
			NULL));
	if (file == INVALID_HANDLE_VALUE) {
		throw IOException(L"CreateFile: " + fileName + L": " + toString(::GetLastError()));
	}
	DWORD length;
	::WriteFile(file, UNICODE_MARK_LITTLE.c_str(), UNICODE_MARK_LITTLE.length() * 2, &length, NULL);
	wstring buffer = out.str();
	::WriteFile(file, buffer.c_str(), buffer.length() * 2, &length, NULL);
}

void RegistryStorage::write(wostream& out) throw(IOException) {
	out << L"Windows Registry Editor Version 5.00" << LN << LN;
	for (	map<wstring, Key*>::const_iterator keys = keyMap.begin();
			keys!= keyMap.end();
			++keys) {
		(*keys).second->write(out);
	}
}


// Key

RegistryStorage::Key::Key(RegistryStorage* storage, const wstring& keyName)
		: storage(storage), keyName(keyName), deleteAtFirst(false) {
}

RegistryStorage::Key::~Key() {
	for (	map<wstring, Value*>::iterator iterator = valueMap.begin();
			iterator != valueMap.end();
			++iterator) {
		delete (*iterator).second;
	}
}

void RegistryStorage::Key::putValue(const wstring& valueName, Value* value) {
	wstring valueNameLow = toLowerCase(valueName);
	Value* oldValue = utils::find_ptr(valueMap, valueNameLow);
	if (oldValue != NULL) {
		delete oldValue;
	}
	valueMap[valueNameLow] = value;
}

void RegistryStorage::Key::setDeleteAtFirst() {
	deleteAtFirst = true;
}

void RegistryStorage::Key::setDWORD(const wstring& valueName, DWORD value) {
	putValue(valueName, new DWORDValue(valueName, value));
}

void RegistryStorage::Key::setBINARY(const wstring& valueName, const BYTE* value, int length) {
	putValue(valueName, new BINARYValue(valueName, value, length));
}

void RegistryStorage::Key::setSZ(const wstring& valueName, const wstring& value) {
	putValue(valueName, new SZValue(valueName, value));
}

void RegistryStorage::Key::setEXPAND_SZ(const wstring& valueName, const wstring& value) {
	putValue(valueName, new EXPAND_SZValue(valueName, value));
}

void RegistryStorage::Key::setMULTI_SZ(const wstring& valueName, const wstring& value) {
	putValue(valueName, new MULTI_SZValue(valueName, value));
}

void RegistryStorage::Key::setString(const wstring& valueName, DWORD type, const wstring& value) {
	switch (type) {
	case REG_SZ:
		setSZ(valueName, value);
		break;
	case REG_EXPAND_SZ:
		setEXPAND_SZ(valueName, value);
		break;
	case REG_MULTI_SZ:
		setMULTI_SZ(valueName, value);
		break;
	default:
		throw IllegalArgumentException(L"type: " + toString(type));
	}
}

void RegistryStorage::Key::setDeleteValue(const wstring& valueName) {
	putValue(valueName, new DeleteValue(valueName));
}

wstring RegistryStorage::Key::getValueString(const wstring& valueName) {
	Value* value = utils::find_ptr(valueMap, toLowerCase(valueName));
	if (value == NULL) return NULL_STRING;
	return value->toString();
}

void RegistryStorage::Key::exportCurrentAll() throw(RegKeyException) {
	exportCurrentValues();

	vector<wstring> subkeys;
	{
		auto_ptr<RegKey> key(new RegKey(keyName, KEY_READ));
		key->getSubKeys(&subkeys);
	}
	for (int i = 0; i < subkeys.size(); i++) {
		wstring subkey = utils::concatPath(keyName, subkeys.at(i));
		storage->getKey(subkey)->exportCurrentAll();
	}
}

void RegistryStorage::Key::exportCurrentValues() throw(RegKeyException) {
	auto_ptr<RegKey> key(new RegKey(keyName, KEY_READ));
	map<wstring, DWORD> types;
	key->getValueTypes(&types);
	for (	map<wstring, DWORD>::const_iterator iterator = types.begin();
			iterator != types.end();
			++iterator) {
		const wstring& name = (*iterator).first;
		DWORD type = (*iterator).second;
		switch (type) {
		case REG_DWORD:
			{
				int valueInt = key->getIntValue(name);
				setDWORD(name, valueInt);
				break;
			}
		case REG_BINARY:
			{
				int length;
				unsigned char* valueBytes = key->getBytesValue(name, &length);
				setBINARY(name, valueBytes, length);
				break;
			}
		case REG_SZ:
		case REG_EXPAND_SZ:
		case REG_MULTI_SZ:
			{
				wstring valueString = key->getStringValue(name);
				setString(name, type, valueString);
				break;
			}
		default:
			throw ConfigException(L"type");
		}
	}
}

void RegistryStorage::Key::write(wostream& out) throw(IOException) {
	if (deleteAtFirst) {
		writeDelete(out);
	}
	out << L"[" << keyName << L"]" << LN;
	for (	map<wstring, Value*>::iterator values = valueMap.begin();
			values != valueMap.end();
			++values) {
		(*values).second->write(out);
	}
	out << LN;
}

void RegistryStorage::Key::writeDelete(wostream& out) throw(IOException) {
	out << L"[-" << keyName << L"]" << LN << LN;
}


// DeleteKey

RegistryStorage::DeleteKey::DeleteKey(RegistryStorage* storage, const wstring& key)
		: Key(storage, key) {
}

void RegistryStorage::DeleteKey::write(wostream& out) throw(IOException) {
	writeDelete(out);
}


// Value

RegistryStorage::Value::Value(const wstring& name) : name(name) {
}

RegistryStorage::Value::~Value() {
}

int RegistryStorage::Value::writeNameEqual(wostream& out) throw(IOException) {
	wstring string;
	if (name == NULL_STRING) {
		string = L"@=";
	} else {
		string = L"\"" + name + L"\"=";
	}
	out << string;
	return string.length();
}


RegistryStorage::DWORDValue::DWORDValue(const wstring& name, DWORD value)
		: Value(name), value(value) {
}

void RegistryStorage::DWORDValue::write(wostream& out) throw(IOException) {
	writeNameEqual(out);
	out		<< L"dword:"
			<< setw(8) << setfill(L'0') << hex << value
			<< LN;
}

wstring RegistryStorage::DWORDValue::toString() {
	return ::toString((int)value);
}


RegistryStorage::AbstractHexValue::AbstractHexValue(const wstring& name, DWORD type, const BYTE* value, int length)
		: Value(name), type(type) {
	this->value = new BYTE[length];
	copy(value, value + length, this->value);
	this->length = length;
}

RegistryStorage::AbstractHexValue::~AbstractHexValue() {
	delete[] value;
}

void RegistryStorage::AbstractHexValue::write(wostream& out) throw(IOException) {
	int xpos = 0;
	xpos += writeNameEqual(out);
	xpos += writeType(out);
	for (int i = 0; i < length; i++) {
		out << setw(2) << setfill(L'0') << hex << (value[i] & 0xff);
		xpos += 2;
		if (i < length - 1) {
			out << L",";
			xpos++;
			if (xpos > 76) {
				out << L"\\" << LN;
				out << L"  ";
				xpos = 2;
			}
		}
	}
	out << LN;
}

int RegistryStorage::AbstractHexValue::writeType(wostream& out) throw(IOException) {
	wstring string = L"hex(" + ::toString((int)type) + L"):";
	out << string;
	return string.length();
}

wstring RegistryStorage::AbstractHexValue::toString() {
	// architecture depended.
	return wstring(reinterpret_cast<const wchar_t*>(value), length / 2);
}


RegistryStorage::BINARYValue::BINARYValue(const wstring& name, const BYTE* value, int length)
		: AbstractHexValue(name, REG_BINARY, value, length) {
}

int RegistryStorage::BINARYValue::writeType(wostream& out) throw(IOException) {
	static const wstring string(L"hex:");
	out << string;
	return string.length();
}


RegistryStorage::SZValue::SZValue(const wstring& name, const wstring& value)
		: Value(name), value(value) {
}

static wstring escape(const wstring& value) {
	int length = value.length();
	wstring buffer;
	buffer.reserve(length);
	for (int i = 0; i < length; i++) {
		wchar_t c = value[i];
		switch (c) {
		case '\\':
			buffer += L"\\\\";
			break;
		case '\"':
			buffer += L"\\\"";
			break;
		default:
			buffer += c;
			break;
		}
	}
	return buffer;
}

void RegistryStorage::SZValue::write(wostream& out) throw(IOException) {
	writeNameEqual(out);
	out << L"\"" << escape(value) << L"\"" << LN;
}

wstring RegistryStorage::SZValue::toString() {
	return value;
}


RegistryStorage::HexStringValue::HexStringValue(const wstring& name, DWORD type, const wstring& value)
		// architecture depended.
		: AbstractHexValue(name, type, reinterpret_cast<const BYTE*>(value.c_str()), (value.length() + 1) * 2),
		value(value) {
}

wstring RegistryStorage::HexStringValue::toString() {
	return value;
}


RegistryStorage::EXPAND_SZValue::EXPAND_SZValue(const wstring& name, const wstring& value)
		: HexStringValue(name, REG_EXPAND_SZ, value) {
}

RegistryStorage::MULTI_SZValue::MULTI_SZValue(const wstring& name, const wstring& value):
		HexStringValue(name, REG_MULTI_SZ, value) {
}

RegistryStorage::DeleteValue::DeleteValue(const wstring& name) : Value(name) {
}

void RegistryStorage::DeleteValue::write(wostream& out) throw(IOException) {
	if (name == NULL_STRING) {
		out << L"@=-";
	} else {
		out << L"\"" << name << L"\"=-";
	}
	out << LN;
}

wstring RegistryStorage::DeleteValue::toString() {
	return L"-";
}
