// tuneapp.


#include "NICEnumerator.h"

#include "../../lang.h"
#include "../../utils.h"


const wstring NICEnumerator::GLOBAL_ID(L"\1global");


NICEnumerator::NICEnumerator() {
}

void NICEnumerator::setParams(xml::Node* params) throw(ConfigException) {
	for (	xml::Node::const_iterator children = params->begin();
			children != params->end();
			++children) {
		xml::Node* child = *children;
		if (child->getType() == xml::Node::ELEMENT) {
			const wstring& name = child->getName();
			if (name == L"label") {
				label = child->getText();
			} else if (name == L"key") {
				regEnumKey = child->getText();
			} else if (name == L"classidValue") {
				regEnumClassidValue = child->getText();
			} else if (name == L"nameValues") {
				setNameValuesParams(child);
			} else if (name == L"global") {
				setGlobalParams(child);
			} else if (name == L"interfaces") {
				setInterfacesParams(child);
			}
		}
	}

	init();
}

void NICEnumerator::setNameValuesParams(xml::Node* params) throw(ConfigException) {
	for (	xml::Node::const_iterator children = params->begin();
			children != params->end();
			++children) {
		xml::Node* child = *children;
		if (child->getType() == xml::Node::ELEMENT) {
			const wstring& name = child->getName();
			if (name == L"nameValue") {
				regEnumNameValues.push_back(child->getText());
			}
		}
	}
}

void NICEnumerator::setGlobalParams(xml::Node* params) throw(ConfigException) {
	NIC global;
	global.id = GLOBAL_ID;
	for (	xml::Node::const_iterator children = params->begin();
			children != params->end();
			++children) {
		xml::Node* child = *children;
		if (child->getType() == xml::Node::ELEMENT) {
			const wstring& name = child->getName();
			if (name == L"name") {
				global.name = child->getText();
			} else if (name == L"key") {
				global.regKey = child->getText();;
			}
		}
	}
	addNIC(global);
}

void NICEnumerator::setInterfacesParams(xml::Node* params) throw(ConfigException) {
	for (	xml::Node::const_iterator children = params->begin();
			children != params->end();
			++children) {
		xml::Node* child = *children;
		if (child->getType() == xml::Node::ELEMENT) {
			const wstring& name = child->getName();
			if (name == L"key") {
				regInterfacesKey = child->getText();
			}
		}
	}
}

void NICEnumerator::init() throw(RegKeyException) {
	getNICs();
}

void NICEnumerator::getNICs() throw(RegKeyException) {
	vector<wstring> subkeys;
	RegKey* rawKeyPtr;
	try {
		rawKeyPtr = new RegKey(regEnumKey, KEY_READ);
	} catch (RegKeyException) {
		return; // no NIC
	}
	{
		auto_ptr<RegKey> key(rawKeyPtr);
		key->getSubKeys(&subkeys);
	}
	for (	vector<wstring>::const_iterator iterator = subkeys.begin();
			iterator != subkeys.end();
			++iterator) {
		try {
			NIC nic;
			wstring subkeyName = utils::concatPath(regEnumKey, *iterator);
			auto_ptr<RegKey> subkey(new RegKey(subkeyName, KEY_READ));
			nic.id = subkey->getStringValue(regEnumClassidValue);
			nic.regKey = utils::concatPath(regInterfacesKey, nic.id);
			delete new RegKey(nic.regKey, KEY_READ);
			nic.name = NULL_STRING;
			for (int j = 0; (nic.name == NULL_STRING) && (j < regEnumNameValues.size()); j++) {
				nic.name = subkey->getStringValue(regEnumNameValues.at(j), NULL_STRING);
			}
			addNIC(nic);
		} catch (RegKeyException) {
		}
	}
}

void NICEnumerator::addNIC(NIC& nic) {
	wstring key = toLowerCase(nic.id);
	keyList.push_back(key);
	nicMap[key] = nic;
}

static bool isGlobalItem(NICParameterItem* item) {
	const wstring& selectionType = item->getSelectionType();
	return	(selectionType == NULL_STRING) ||
			(selectionType == NICParameterItem::SELTYPE_GLOBAL_ONLY);
}

static bool isInterfacesItem(NICParameterItem* item) {
	const wstring& selectionType = item->getSelectionType();
	return	(selectionType == NULL_STRING) ||
			(selectionType == NICParameterItem::SELTYPE_INTERFACES_ONLY);
}

void NICEnumerator::addGroup(ItemGroup* templateGroup) {
	int i = 0;
	for (	vector<wstring>::const_iterator keys = keyList.begin();
			keys != keyList.end();
			++keys, i++) {
		const wstring& key = *keys;
		const NIC& nic = nicMap[key];
		ItemGroup* group = new ItemGroup();
		group->setParent(this);
		group->setID(templateGroup->getID() + L"." + key);
		group->setName(templateGroup->getName());
		group->setHelpText(templateGroup->getHelpText());
		for (	ItemGroup::const_iterator items = templateGroup->begin();
				items != templateGroup->end();
				++items) {
			Item* item = *items;
			if (NICParameterItem* templateItem = dynamic_cast<NICParameterItem*>(item)) {
				if ((i == 0) && isGlobalItem(templateItem)) {
					item = templateItem;
				} else if ((i > 0) && isInterfacesItem(templateItem)) {
					item = templateItem->cloneForInterface(nic.regKey);
				} else {
					item = NULL;
				}
				if (item != NULL) {
					item->setParent(group);
					group->addItem(item);
				}
			} else {
				if (i == 0) {
					group->addItem(item);
				}
			}
		}
		if (group->countItems() > 0) {
			ItemGroupSelector::addGroup(group);
			nicMap[key].groupList.push_back(group);
		}
	}
}

const vector<wstring>& NICEnumerator::keys() const {
	return keyList;
}

const wstring& NICEnumerator::getLabel() const {
	return label;
}

const wstring& NICEnumerator::getLabel(const wstring& key) const {
	map<wstring, NIC>::const_iterator found = nicMap.find(key);
	if (found == nicMap.end()) {
		throw IllegalArgumentException(key);
	}
	return (*found).second.name;
}

const ItemGroupList& NICEnumerator::getItemGroupList(const wstring& key) const {
	map<wstring, NIC>::const_iterator found = nicMap.find(key);
	if (found == nicMap.end()) {
		throw IllegalArgumentException(key);
	}
	return (*found).second.groupList;
}

const wstring& NICEnumerator::getShortHelpText() const {
	// Not Implemented.
	// return guid.
	return NULL_STRING;
}
