// tuneapp.


#include "utils.h"
#include "types.h"


namespace utils {


awt::FlowLayout* leftFlowLayout;
awt::GridLayout* rowsGridLayout;
awt::GridFlowLayout* c2GridFlowLayout;


void init() {
	leftFlowLayout = new awt::FlowLayout(awt::FlowLayout::LEFT);
	rowsGridLayout = new awt::GridLayout(0, 1);
	c2GridFlowLayout = new awt::GridFlowLayout(0, 2);
	c2GridFlowLayout->setHalign(awt::GridFlowLayout::LEFT);
	c2GridFlowLayout->setCellValign(awt::GridFlowLayout::CENTER);
}


wstring getIDPath(const Item* item) {
	ItemGroup* group = item->getParent();
	Category* category = group->getParent();
	return category->getID() + L"." + group->getID() + L"." + item->getID();
}


int hexToInt(wchar_t c) throw(NumberFormatException) {
	if ((c >= '0') && (c <= '9')) {
		return c - '0';
	} else if ((c >= 'A') && (c <= 'F')) {
		return c - 'A' + 10;
	} else if ((c >= 'a') && (c <= 'f')) {
		return c - 'a' + 10;
	} else {
		throw NumberFormatException(L"hexToInt: " + c);
	}
}

static const char const HEX_TABLE[] = {
	'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
	'a', 'b', 'c', 'd', 'e', 'f'
};

wstring byteToHexString(signed char b) {
	wstring string;
	string.reserve(2);
	string += HEX_TABLE[(b & 0xf0) >> 4];
	string += HEX_TABLE[(b & 0x0f)     ];
	return string;
}


wstring concatPath(const wstring& parent, const wstring& child) {
	if (parent.empty()) {
		return child;
	} else if (child.empty()) {
		return parent;
	} else {
		if (parent[parent.length() - 1] == '\\') {
			return parent + child;
		} else {
			return parent + L"\\" + child;
		}
	}
}

wstring getAbsolutePath(const wstring& file) {
	auto_array<wchar_t> buffer(_MAX_PATH + 1);
	wchar_t* filePart;
	if (::GetFullPathName(file.c_str(), _MAX_PATH + 1, buffer, &filePart) == 0) {
		throw;
	}
	return wstring(buffer);
}

wstring getFileBaseName(const wstring& path) {
	wstring::size_type separator = path.rfind('\\');
	if (separator == wstring::npos) {
		return path;
	} else {
		return path.substr(separator + 1);
	}
}

wstring getFileBasePath(const wstring& path) {
	wstring::size_type separator = path.rfind('\\');
	if ((separator == wstring::npos) || (separator == 0)) {
		return path;
	} else {
		return path.substr(0, separator);
	}
}


int readFileAll(const wstring& fileName, wstring* buffer) throw(IOException) {
	auto_handle file(::CreateFile(
			fileName.c_str(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL));
	if (file == INVALID_HANDLE_VALUE) {
		DWORD error = ::GetLastError();
		if (error == ERROR_FILE_NOT_FOUND) {
			throw FileNotFoundException(fileName);
		}
		throw IOException(L"Failed to open: " + fileName + L" : " + toString(::GetLastError()));
	}
	DWORD size = ::GetFileSize(file, NULL);
	if (size == -1) {
		throw IOException(L"Failed to get file size: " + fileName + L" : " + toString(::GetLastError()));
	}
	if ((size < 2) || ((size & 1) == 1)) {
		throw IOException(L"No unicode file: " + fileName);
	}
	wchar_t mark;
	DWORD read;
	BOOL result = ::ReadFile(file, &mark, sizeof(wchar_t), &read, NULL);
	if ((result == FALSE) || (read != sizeof(wchar_t))) {
		throw IOException(L"Failed to read: " + fileName + L" : " + toString(::GetLastError()));
	}
	if (mark != 0xfeff) {
		throw IOException(L"No unicode little file: " + fileName);
	}
	size -= 2;
	auto_array<BYTE> bytes(size);
	result = ::ReadFile(file, bytes, size, &read, NULL);
	if ((result == FALSE) || (read != size)) {
		throw IOException(L"Failed to read: " + fileName + L" : " + toString(::GetLastError()));
	}
	// architecture depended: byte order
	buffer->append(reinterpret_cast<wchar_t*>(bytes.ptr()), size / 2);
	return size / 2;
}


wstring getWindowsDirectory() {
	UINT length = MAX_PATH + 1;
	auto_array<wchar_t> dir(length);
	length = ::GetWindowsDirectory(dir, length);
	if (length == 0) {
		throw IOException(L"Can't get windows directory: " + toString(::GetLastError()));
	}
	return wstring(dir, length);
}

wstring getSystemDirectory() {
	UINT length = MAX_PATH + 1;
	auto_array<wchar_t> dir(length);
	length = ::GetSystemDirectory(dir, length);
	if (length == 0) {
		throw IOException(L"Can't get system directory: " + toString(::GetLastError()));
	}
	return wstring(dir, length);
}


void exec(const wstring& command, const wstring& args) throw(IOException) {
	STARTUPINFO startInfo;
	PROCESS_INFORMATION processInfo;
	::ZeroMemory(&startInfo, sizeof(STARTUPINFO));
	startInfo.cb = sizeof(STARTUPINFO);
	BOOL result = ::CreateProcess(
			command.c_str(), const_cast<wchar_t*>(args.c_str()),
			NULL, NULL, FALSE, CREATE_DEFAULT_ERROR_MODE | CREATE_NO_WINDOW, NULL, NULL,
			&startInfo, &processInfo);
	if (result == FALSE) {
		throw IOException(L"CreateProcess(" + command + L") failed: " + toString(::GetLastError()));
	}
	::CloseHandle(processInfo.hThread);
	::CloseHandle(processInfo.hProcess);
}


} // namespace
