// Rick DOM.


#include "Parser.h"
#include "../types.h"

#include <assert.h>


namespace xml {


Parser::Parser(ParserCallback* callback)
		: callback(callback) {
}

void Parser::parse(std::wistream& in) throw(XMLException) {
	readXmlDeclaration(in);
	readElements(in);
	callback->endDocument();
}


static bool isWhiteSpace(wchar_t c) {
	return (c <= 0x0020);
}

static bool isQuotationChar(wchar_t c) {
	return ((c == '\'') || (c == '\"'));
}

void Parser::readXmlDeclaration(std::wistream& in) {
	std::wstring name;
	Attributes attrs;
	std::wstring extra;
	readTagLikeStructure(in, name, attrs, extra);
	if (name != L"?xml") {
		throw XMLException(L"no xml declaration");
	}
	const std::wstring& version = attrs[L"version"];
	if (version != L"1.0") {
		throw XMLException(L"parser supports xml version 1.0 only, but specified document version: " + version);
	}
}

void Parser::readElements(std::wistream& in) {
	while (true) {
		std::wistream::int_type c = skipWhiteSpace(in);
		if (c == WEOF) break;
		if (c == '<') {
			readTagLikeStructureAfterLt(in);
		} else {
			readTextAfter(in, c);
		}
	}
}

std::wistream::int_type Parser::skipWhiteSpace(std::wistream& in) {
	while (true) {
		std::wistream::int_type c = in.get();
		if ((c == WEOF) || !isWhiteSpace(c)) {
			return c;
		}
	}
}

Parser::Type Parser::readTagLikeStructure(std::wistream& in, std::wstring& name, Attributes& attrs, std::wstring& extra) {
	std::wistream::int_type c = in.get();
	if (c == WEOF) {
		throw XMLException(L"unexpected EOF");
	} else if (c != '<') {
		throw XMLException(L"no tag like structure");
	}
	return readTagLikeStructureAfterLt(in, name, attrs, extra);
}

void Parser::readTagLikeStructureAfterLt(std::wistream& in) {
	std::wstring name;
	Attributes attrs;
	std::wstring extra;
	Type type = readTagLikeStructureAfterLt(in, name, attrs, extra);
	switch (type) {
	case START_TAG:
		callback->startElement(name, attrs);
		break;
	case END_TAG:
		callback->endElement(name.substr(1));
		break;
	case EMPTY_TAG:
		callback->startElement(name, attrs);
		callback->endElement(name);
		break;
	case COMMENT:
		// callback->comment(extra);
		break;
	case PROCESSING_INSTRUCTION:
		// callback->instruction(name);
		break;
	case CDATA_SECTION:
		callback->text(extra);
		break;
	case SECTION:
		// callback->section(name);
		break;
	default:
		assert (false);
	}
}

Parser::Type Parser::readTagLikeStructureAfterLt(std::wistream& in, std::wstring& name, Attributes& attrs, std::wstring& extra) {
	std::wistream::int_type c = in.get();
	if (c == WEOF) {
		throw XMLException(L"unexpected EOF");
	} else if (c == '?') {
		readProcessingInstructionAfterQ(in, name, attrs);
		return PROCESSING_INSTRUCTION;
	} else if (c == '!') {
		return readSectionLikeStructureAfterEx(in, name, attrs, extra);
	} else if (isWhiteSpace(c)) {
		throw XMLException(L"needs name after \'<\'");
	} else {
		bool empty = true;
		readTagAfter(in, c, name, attrs, L">", empty);
		assert (!name.empty());
		if (name[0] == '/') return END_TAG;
		else if (empty) return EMPTY_TAG;
		else return START_TAG;
	}
}

void Parser::readProcessingInstructionAfterQ(std::wistream& in, std::wstring& name, Attributes& attrs) {
	bool empty = false;
	readTagAfter(in, '?', name, attrs, L"?>", empty);
}

Parser::Type Parser::readSectionLikeStructureAfterEx(std::wistream& in, std::wstring& name, Attributes& attrs, std::wstring& extra) {
	std::wistream::int_type c = in.get();
	if (c == WEOF) {
		throw XMLException(L"unexpected EOF");
	} else if (c == '-') {
		readCommentAfterHyph(in, c, name, extra);
		return COMMENT;
	} else if (c == '[') {
		return readConditionalSectionAfterBracket(in, name, extra);
	} else {
		readSectionAfter(in, c, name, extra);
		return SECTION;
	}
}

void Parser::readTagAfter(std::wistream& in, wchar_t preC, std::wstring& name, Attributes& attrs, const std::wstring& endMark, bool& empty) {
	name += preC;
	while (true) {
		std::wistream::int_type c = in.get();
		if (c == WEOF) {
			throw XMLException(L"unexpected EOF");
		} else if (isWhiteSpace(c)) {
			if (name.empty()) {
				throw XMLException(L"needs name after \'<\'");
			}
			readAttributes(in, attrs, endMark, empty);
			break;
		} else if (c == endMark.at(0)) {
			if (name.empty()) {
				throw XMLException(L"needs name after \'<\'");
			}
			perhapsTagEnd(in, c, endMark);
			empty = false;
			break;
		} else if (empty && (c == '/')) {
			if (name.empty()) {
				throw XMLException(L"needs name after \'<\'");
			}
			c = in.get();
			if (c == WEOF) {
				throw XMLException(L"unexpected EOF");
			} else if (c == endMark[0]) {
				perhapsTagEnd(in, c, endMark);
				// empty = true;
				break;
			} else {
				throw XMLException(L"illegal character in tag: \'/\'");
			}
		} else {
			name += wchar_t(c);
		}
	}
}

void Parser::readAttributes(std::wistream& in, Attributes& attrs, const std::wstring& endMark, bool& empty) {
	while (true) {
		std::wstring attr;
		std::wstring value;
		bool tagOut = readAttribute(in, attr, value, endMark, empty);
		if (!attr.empty()) {
			attrs[attr] = value;
		}
		if (tagOut) break;
	}
}

bool Parser::readAttribute(std::wistream& in, std::wstring& attrName, std::wstring& value, const std::wstring& endMark, bool& empty) {
	std::wistream::int_type c;
	while (true) {
		c = in.get();
		if (c == WEOF) {
			throw XMLException(L"unexpected EOF");
		}
		if (!isWhiteSpace(c)) break;
	}
	while (true) {
		if (c == '=') {
			if (attrName.empty()) {
				throw XMLException(L"\'=\' without attribute name");
			}
			return readAttributeValue(in, value, endMark, empty);
		} else if (isWhiteSpace(c)) {
			assert (!attrName.empty());
			throw XMLException(L"no attribute value for: " + attrName);
			return false;
		} else if (c == endMark[0]) {
			if (!attrName.empty()) {
				throw XMLException(L"no attribute value for: " + attrName);
			}
			perhapsTagEnd(in, c, endMark);
			empty = false;
			return true;
		} else if (empty && (c == '/')) {
			if (!attrName.empty()) {
				throw XMLException(L"no attribute value for: " + attrName);
			}
			c = in.get();
			if (c == WEOF) {
				throw XMLException(L"unexpected EOF");
			} else if (c == endMark[0]) {
				perhapsTagEnd(in, c, endMark);
				// empty = true;
				return true;
			} else {
				throw XMLException(L"illegal character in tag: \'/\'");
			}
		} else {
			attrName += wchar_t(c);
		}
		c = in.get();
		if (c == WEOF) {
			throw XMLException(L"unexpected EOF");
		}
	}
}

bool Parser::readAttributeValue(std::wistream& in, std::wstring& value, const std::wstring& endMark, bool& empty) {
	std::wistream::int_type quotation = 0;
	while (true) {
		std::wistream::int_type c = in.get();
		if (!quotation) {
			if (isWhiteSpace(c)) {
				return false;
			} else if (c == endMark.at(0)) {
				perhapsTagEnd(in, c, endMark);
				empty = false;
				return true;
			} else if (empty && (c == '/')) {
				c = in.get();
				if (c == endMark.at(0)) {
					perhapsTagEnd(in, c, endMark);
					// emptry = true;
					return true;
				} else {
					throw XMLException(L"illegal character in tag: \'/\'");
				}
			} else if (isQuotationChar(c)) {
				if (!value.empty()) {
					throw XMLException(L"invalid quotation");
				}
				quotation = c;
			} else {
				throw XMLException(L"unquoted attribute value");
			}
		} else {
			if (c == quotation) {
				quotation = 0;
			} else if (c == '&') {
				readReferenceAfterAmp(in, value);
			} else {
				value += wchar_t(c);
			}
		}
	}
}

void Parser::perhapsTagEnd(std::wistream& in, wchar_t preC, const std::wstring& endMark) {
	assert (preC == endMark[0]);
	for (int i = 1; i < endMark.length(); i++) {
		std::wistream::int_type c = in.get();
		if (c == WEOF) {
			throw XMLException(L"unexpected EOF");
		} else if (c != endMark[i]) {
			throw XMLException(L"no end mark " + endMark);
		}
	}
}

void Parser::readCommentAfterHyph(std::wistream& in, wchar_t preC, std::wstring& name, std::wstring& extra) {
	std::wistream::int_type c = in.get();
	if (c != '-') {
		throw XMLException(L"invalid section <!-" + wchar_t(c));
	}
	name = L"!--";
	readUntil(in, extra, L"-->");
}

Parser::Type Parser::readConditionalSectionAfterBracket(std::wistream& in, std::wstring& name, std::wstring& extra) {
	name = L"![";
	while (true) {
		std::wistream::int_type c = in.get();
		if (c == WEOF) {
			throw XMLException(L"unexpected EOF");
		}
		name += wchar_t(c);
		if (c == '[') break;
	}
	if (name == L"![CDATA[") {
		readCdataSection(in, extra);
		return CDATA_SECTION;
	} else {
		readUntil(in, extra, L"]]>");
		return SECTION;
	}
}

void Parser::readCdataSection(std::wistream& in, std::wstring& extra) {
	static const std::wstring endMark = L"]]>";
	int match = 0;
	while (true) {
		std::wistream::int_type c = in.get();
		if (c == WEOF) {
			throw XMLException(L"unexpected EOF");
		} else if (c == endMark[match]) {
			for (match = 1; match < endMark.length(); ++match) {
				c = in.get();
				if (c == WEOF) {
					throw XMLException(L"unexpected EOF");
				}
				if (c != endMark[match]) break;
			}
			if (match == endMark.length()) break;
			extra.append(endMark, 0, match);
			match = 0;
		} else {
			extra += wchar_t(c);
		}
	}
}

void Parser::readSectionAfter(std::wistream& in, wchar_t preC, std::wstring& name, std::wstring& extra) {
	// Not Implemented
	name = L"!";
	std::wistream::int_type c = preC;
	while (true) {
		if (isWhiteSpace(c)) {
			break;
		} else if (c == '>') {
			return;
		}
		name += wchar_t(c);
		c = in.get();
		if (c == WEOF) {
			throw XMLException(L"unexpected EOF");
		}
	}
	readUntil(in, extra, L">");
}

void Parser::readUntil(std::wistream& in, std::wstring& extra, const std::wstring& endMark) {
	int match = 0;
	while (true) {
		std::wistream::int_type c = in.get();
		if (c == WEOF) {
			throw XMLException(L"unexpected EOF");
		} else if (c == endMark[match]) {
			for (match = 1; match < endMark.length(); ++match) {
				c = in.get();
				if (c == WEOF) {
					throw XMLException(L"unexpected EOF");
				}
				if (c != endMark[match]) break;
			}
			if (match == endMark.length()) break;
			extra.append(endMark, 0, match);
			match = 0;
		} else if (c == '&') {
			readReferenceAfterAmp(in, extra);
		} else {
			extra += wchar_t(c);
		}
	}
}

void Parser::readTextAfter(std::wistream& in, wchar_t preC) {
	assert (preC != '<');
	std::wstring text;
	std::wistream::int_type c = preC;
	while (true) {
		if (c == '<') {
			in.unget();
			break;
		} else if (c == '&') {
			readReferenceAfterAmp(in, text);
		} else {
			text += wchar_t(c);
		}
		c = in.get();
		if (c == WEOF) break;
	}
	callback->text(text);
}

void Parser::readReferenceAfterAmp(std::wistream& in, std::wstring& dest) {
	std::wistream::int_type c = in.get();
	if (c == WEOF) {
		throw XMLException(L"unexpected EOF");
	} else if (c == '#') {
		dest += readNumericReferenceAfterSharp(in);
	} else {
		readEntityReferenceAfter(in, c, dest);
	}
}

wchar_t Parser::readNumericReferenceAfterSharp(std::wistream in) {
	int number = 0;
	bool hex = false;
	for (int length = 0; ; length++) {
		std::wistream::int_type c = in.get();
		if (c == WEOF) {
			throw XMLException(L"unexpected EOF");
		} else if (c == ';') {
			if (length == 0) {
				throw XMLException(L"invalid reference: &#;");
			}
			return number;
		} else if (c == 'x') {
			if (length > 0) {
				throw XMLException(L"invalid character in numeric reference: x");
			}
			hex = true;
		} else {
			if ((c >= '0') && (c <= '9')) {
				number = number * 10 + (c - '0');
			} else if (hex) {
				if ((c >= 'A') && (c <= 'F')) {
					number = number * 16 + (c - 'A' + 10);
				} else if ((c >= 'a') && (c <= 'f')) {
					number = number * 16 + (c - 'a' + 10);
				} else {
					throw XMLException(L"invalid character in numeric reference: " + wchar_t(c));
				}
			} else {
				throw XMLException(L"invalid character in numeric reference: " + wchar_t(c));
			}
			// check
			if (number == 0) {
				throw XMLException(L"numeric reference starts 0");
			} else if (number > 0xffff) {
				throw XMLException(L"numeric reference so big");
			}
		}
	}
}

void Parser::readEntityReferenceAfter(std::wistream& in, wchar_t preC, std::wstring& dest) {
	std::wstring name;
	std::wistream::int_type c = preC;
	while (true) {
		if (c == ';') {
			const std::wstring& value = entityTable.get(name);
			if (value == NULL_STRING) {
				throw XMLException(L"unknown entity name refererenced: &" + name + L";");
			}
			dest += value;
			break;
		} else {
			name += wchar_t(c);
		}
		c = in.get();
		if (c == WEOF) {
			throw XMLException(L"unexpected EOF");
		}
	}
}


} // namespace
